# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" Form validators for Pyramid """


import datetime
import re

import formencode
from formencode import validators
from formencode.compound import from_python
from formencode.national import InternationalPhoneNumber

from pyramid_helpers.i18n import N_


COLOR_RE = re.compile(r'#(?P<r>[0-9a-f]{2})(?P<v>[0-9a-f]{2})(?P<b>[0-9a-f]{2})', re.I)
MONTH_RE = re.compile(r'(?P<year>\d{4})-(?P<month>\d{2})')
TIME_RE = re.compile(r'(?P<hours>\d{2}):(?P<minutes>\d{2})')
WEEK_RE = re.compile(r'(?P<year>\d{4})-W(?P<week>\d{2})')


#
# Validators
#
class Color(formencode.FancyValidator):
    """ Formencode validator for color representation (#RRVVBB) """

    messages = {
        'invalid': N_('Invalid value'),
    }

    format = 'hex'

    # pylint: disable=unused-argument, no-self-use
    def _convert_from_python(self, value, state):
        if isinstance(value, (list, tuple)):
            value = '#{0:02x}{1:02x}{2:02x}'.format(*value)

        return value

    def _convert_to_python(self, value, state):
        match = COLOR_RE.match(value)
        if match is None:
            raise formencode.Invalid(self.message('invalid', state), value, state)

        if self.format == 'hex':
            return value

        return [int(v, 16) for v in match.groups()]


class DateTime(validators.FancyValidator):
    """ Formencode validator for datetime representation using given format """

    messages = {
        'invalid': N_('Invalid value'),
        'year': N_('Year must be >= 1900'),
    }

    format = N_('%m/%d/%Y')
    is_date = False
    is_localized = True
    is_naive_utc = False

    def _convert_from_python(self, value, state):
        if not isinstance(value, (datetime.datetime, datetime.date)):
            raise formencode.Invalid(self.message('invalid', state), value, state)

        format_ = state.translate(self.format) if self.is_localized else self.format

        # strftime requires year to be >= 1900
        if value.year < 1900:
            raise formencode.Invalid(self.message('year', state), value, state)

        if not self.is_date and self.is_naive_utc:
            # Convert naive utc to local
            value = state.request.utctolocal(value)

        return value.strftime(format_)

    def _convert_to_python(self, value, state):
        if isinstance(value, (datetime.datetime, datetime.date)):
            return value

        format_ = state.translate(self.format) if self.is_localized else self.format
        try:
            date = datetime.datetime.strptime(value, format_)
            if self.is_date:
                date = date.date()
            elif self.is_naive_utc:
                # Convert naive local to utc
                date = state.request.localtoutc(date)
        except ValueError as exc:
            raise formencode.Invalid(self.message('invalid', state), value, state) from exc

        # strftime used in _convert_from_python requires year to be >= 1900
        if date.year < 1900:
            raise formencode.Invalid(self.message('year', state), value, state)

        return date


class List(formencode.ForEach):
    """
    Modified version of formencode.ForEach which uses csv format in a single
    field instead of repeating the field.
    """

    messages = {
        'invalid': N_('Invalid value'),
    }

    format = '{}'
    if_empty = []
    if_missing = []
    separator = ','
    strip = True

    # pylint: disable=comparison-with-callable
    def _attempt_convert(self, value, state, validate):
        if validate == from_python:
            # Calling inherited
            value = super()._attempt_convert(value, state, validate)

            try:
                value = self.separator.join(self.format.format(v) for v in value)
            except Exception as exc:
                raise formencode.Invalid(self.message('invalid', state), value, state) from exc
        else:
            # to_python
            try:
                value = value.split(self.separator)

                if self.strip:
                    value = [v.strip() for v in value]

            except Exception as exc:
                raise formencode.Invalid(self.message('invalid', state), value, state) from exc

            # Calling inherited
            value = super()._attempt_convert(value, state, validate)

        return value

    def empty_value(self, value):
        return ''


class Month(formencode.FancyValidator):
    """ Formencode validator for month representation (YYYY-MM) """

    messages = {
        'invalid': N_('Invalid value'),
        'month-min': N_('Month must be >= 1'),
        'month-max': N_('Month must be <= 12'),
        'year-min': N_('Year must be >= 1900')
    }

    # pylint: disable=unused-argument, no-self-use
    def _convert_from_python(self, value, state):
        if isinstance(value, (list, tuple)):
            value = '{0:04d}-{1:02d}'.format(*value)

        return value

    def _convert_to_python(self, value, state):
        match = MONTH_RE.match(value)
        if match is None:
            raise formencode.Invalid(self.message('invalid', state), value, state)

        year, month = list(map(int, match.groups()))

        if year < 1900:
            raise formencode.Invalid(self.message('year-min', state), value, state)

        if month < 1:
            raise formencode.Invalid(self.message('month-min', state), value, state)

        if month > 12:
            raise formencode.Invalid(self.message('month-max', state), value, state)

        return (year, month)


class Time(formencode.FancyValidator):
    """ Formencode validator for time representation (hh:mm) """

    messages = {
        'invalid': N_('Invalid value'),
        'hours-min': N_('Hours must be >= 00'),
        'hours-max': N_('Hours must be <= 23'),
        'minutes-min': N_('Minutes must be >= 00'),
        'minutes-max': N_('Minutes must be <= 59'),
    }

    # pylint: disable=unused-argument, no-self-use
    def _convert_from_python(self, value, state):
        if isinstance(value, (list, tuple)):
            value = '{0:02d}:{1:02d}'.format(*value)

        return value

    def _convert_to_python(self, value, state):
        match = TIME_RE.match(value)
        if match is None:
            raise formencode.Invalid(self.message('invalid', state), value, state)

        hours, minutes = list(map(int, match.groups()))

        if hours < 0:
            raise formencode.Invalid(self.message('hours-min', state), value, state)

        if hours > 23:
            raise formencode.Invalid(self.message('hours-max', state), value, state)

        if minutes < 0:
            raise formencode.Invalid(self.message('minutes-min', state), value, state)

        if minutes > 59:
            raise formencode.Invalid(self.message('minutes-max', state), value, state)

        return (hours, minutes)


class Week(formencode.FancyValidator):
    """ Formencode validator for week representation (YYYY-WXX) """

    messages = {
        'invalid': N_('Invalid value'),
        'week-min': N_('Week must be >= 1'),
        'week-max': N_('Week must be <= 52'),
        'year-min': N_('Year must be >= 1900'),
    }

    # pylint: disable=unused-argument, no-self-use
    def _convert_from_python(self, value, state):
        if isinstance(value, (list, tuple)):
            value = '{0:04d}-W{1:02d}'.format(*value)

        return value

    def _convert_to_python(self, value, state):
        match = WEEK_RE.match(value)
        if match is None:
            raise formencode.Invalid(self.message('invalid', state), value, state)

        year, week = list(map(int, match.groups()))

        if year < 1900:
            raise formencode.Invalid(self.message('year-min', state), value, state)

        if week < 1:
            raise formencode.Invalid(self.message('week-min', state), value, state)

        if week > 52:
            raise formencode.Invalid(self.message('week-max', state), value, state)

        return (year, week)


#
# Test Form
#
class ValidatorsForm(formencode.Schema):
    """ Pyramid-Helpers test form """

    allow_extra_fields = True
    filter_extra_fields = True

    checkbox_input = validators.StringBool(if_missing=False, if_empty=False)
    color_input = Color(format='rvb', if_missing=None, if_empty=None)
    date_input = DateTime(format='%Y-%m-%d', if_missing=None, if_empty=None, is_date=True, is_localized=False)
    datetime_local_input = DateTime(format='%Y-%m-%dT%H:%M', if_missing=None, if_empty=None, is_localized=False)
    datetime_naive_input = DateTime(format='%Y-%m-%dT%H:%M', if_missing=None, if_empty=None, is_naive_utc=True)
    email_input = validators.Email(if_missing=None, if_empty=None)
    list_input = List()
    month_input = Month(if_missing=None, if_empty=None)
    number_input = validators.Number(if_missing=None, if_empty=None)
    password_input = validators.String(if_missing=None, if_empty=None)
    radio_input = validators.OneOf(['one', 'two', 'three'], if_missing=None, if_empty=None)
    range_input = validators.Int(min=0, max=100, if_missing=None, if_empty=None)
    search_input = validators.String(if_missing=None, if_empty=None)
    select_input1 = validators.OneOf(['one', 'two', 'three'], if_missing=None, if_empty=None)
    select_input2 = validators.OneOf(['banana', 'orange', 'apple'], if_missing=None, if_empty=None)
    tel_input = InternationalPhoneNumber(if_missing=None, if_empty=None)
    text_input = validators.String(if_missing=None, if_empty=None)
    textarea_input = validators.String(if_missing=None, if_empty=None)
    time_input = Time(if_missing=None, if_empty=None)
    upload_input = validators.FieldStorageUploadConverter(if_missing=None, if_empty=None)
    url_input = validators.URL(if_missing=None, if_empty=None)
    week_input = Week(if_missing=None, if_empty=None)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers common views """

import hashlib

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.httpexceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPInternalServerError
from pyramid.httpexceptions import HTTPNotFound
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.security import forget
from pyramid.security import remember
from pyramid.view import exception_view_config
from pyramid.view import view_config

from pyramid_helpers.api_doc import api_doc as api_doc_
from pyramid_helpers.auth import check_credentials
from pyramid_helpers.forms import validate
from pyramid_helpers.forms.auth import LoginForm
from pyramid_helpers.forms.validators import ValidatorsForm
from pyramid_helpers.i18n import N_
from pyramid_helpers.utils import get_settings


# HTTPExceptions' titles
N_('Bad Request')
N_('Forbidden')
N_('Not Found')

N_('Access was denied to this resource.')
N_('The resource could not be found.')


@view_config(route_name='api-doc', renderer='')
def api_doc(request):
    """ The API doc view """

    translate = request.translate

    renderer_values = dict(
        breadcrumb=[(translate('API documentation'), None)],
        subtitle=None,
        title=translate('API documentation'),
    )

    return api_doc_(request, renderer_values)


@exception_view_config(HTTPBadRequest, renderer='/errors.mako')
def http_400(request):
    """ HTTP 400 error view """

    translate = request.translate

    request.response.status_int = 400

    if request.matched_route.name.startswith('api.'):
        request.override_renderer = 'json'
        result = dict(
            url=request.url,
            method=request.method,
            params=request.params.mixed(),
            errors=request.exception.json if request.exception.body else {},
            apiVersion='1.0',
            result=False,
            message=request.exception.message or translate('Invalid or missing parameter(s)'),
        )
        return result

    # Prevent from rendering wrong template
    if hasattr(request, 'override_renderer'):
        del request.override_renderer

    return dict(
        title=translate(request.exception.title),
        subtitle=None,
    )


@exception_view_config(HTTPForbidden, renderer='/errors.mako')
@exception_view_config(HTTPUnauthorized, renderer='/errors.mako')
def http_403(request):
    """ HTTP 403 error view """

    translate = request.translate

    if request.authenticated_user is None:
        params = get_settings(request, 'auth', 'auth')
        if params.get('policies') == ['basic']:
            # Issuing a challenge
            response = HTTPUnauthorized()
            response.headers.update(forget(request))
            return response

        return HTTPFound(location=request.route_path('login', _query=dict(redirect=request.url)))

    request.response.status_int = 403

    if request.matched_route.name.startswith('api.'):
        request.override_renderer = 'json'
        result = dict(
            url=request.url,
            method=request.method,
            params=request.params.mixed(),
            apiVersion='1.0',
            result=False,
            message=request.exception.message or translate('Access denied'),
        )
        return result

    # Prevent from rendering wrong template
    if hasattr(request, 'override_renderer'):
        del request.override_renderer

    return dict(
        title=translate(request.exception.title),
        subtitle='',
    )


@exception_view_config(HTTPNotFound, renderer='/errors.mako')
def http_404(request):
    """ HTTP 404 error view """

    translate = request.translate

    request.response.status_int = 404

    if request.matched_route and request.matched_route.name.startswith('api.'):
        request.override_renderer = 'json'
        result = dict(
            url=request.url,
            method=request.method,
            params=request.params.mixed(),
            apiVersion='1.0',
            result=False,
            message=request.exception.message or translate('Not found'),
        )
        return result

    # Prevent from rendering wrong template
    if hasattr(request, 'override_renderer'):
        del request.override_renderer

    return dict(
        title=translate(request.exception.title),
        subtitle='',
    )


@view_config(route_name='index', renderer='index.mako')
def index(request):
    """ Index view """

    return dict(
        breadcrumb=[],
        title='Pyramid Helpers',
        subtitle='',
    )


@view_config(route_name='i18n', renderer='/i18n.mako')
def i18n(request):
    """ I18n view """

    translate = request.translate

    return dict(
        breadcrumb=[
            (translate('I18n'), None),
        ],
        title=translate('I18n page'),
        subtitle='',
    )


@view_config(route_name='login', renderer='/login.mako')
@validate('login_form', LoginForm)
def login(request):
    """ Login view """

    translate = request.translate
    form = request.forms['login_form']

    params = get_settings(request, 'auth', 'auth')
    if params.get('policies') == ['remote']:
        params = get_settings(request, 'auth', 'policy:remote')
        login_url = params.get('login_url')
        if login_url is None:
            raise HTTPInternalServerError(detail=translate('Authentication not available, please retry later.'))

        return HTTPFound(location=login_url)

    if request.method == 'POST':
        if not form.errors:
            username = form.result['username']
            password = form.result['password']
            redirect = form.result['redirect']

            if check_credentials(username, password, request):
                headers = remember(request, username)
                return HTTPFound(location=redirect, headers=headers)

            form.errors['username'] = translate('Bad user or password')
    else:
        redirect = request.GET.get('redirect')
        if redirect is None or redirect == request.current_route_url(_query=None):
            redirect = request.route_path('index')

        data = dict(redirect=redirect)
        form.from_python(data)

    return dict(
        title=translate('Authentication'),
    )


@view_config(route_name='logout', renderer='/logout.mako')
def logout(request):
    """ Logout view """

    session = request.session
    translate = request.translate

    # Clear session
    session.delete()

    if request.authentication_policy == 'remote':
        params = get_settings(request, 'auth', 'policy:remote')
        logout_url = params.get('logout_url')
        if logout_url is None:
            # Display logout page
            return dict(
                title=translate('Logged out'),
            )
    else:
        logout_url = request.route_path('index')

    headers = forget(request)
    return HTTPFound(location=logout_url, headers=headers)


@view_config(route_name='predicates', renderer='/predicates.mako')
@view_config(route_name='predicates.enum', renderer='/predicates.mako')
@view_config(route_name='predicates.numeric-1', renderer='/predicates.mako')
@view_config(route_name='predicates.numeric-2', renderer='/predicates.mako')
def predicates(request):
    """ Predicates view """

    translate = request.translate

    return dict(
        breadcrumb=[
            (translate('Predicates'), None),
        ],
        title=translate('Predicates page'),
        subtitle='',
    )


@view_config(route_name='validators', renderer='/validators.mako')
@validate('validators_form', ValidatorsForm, extract='post')
def validators(request):
    """ Validators view """

    translate = request.translate
    form = request.forms['validators_form']

    if request.method == 'POST':
        if not form.errors:
            qfile = form.result['upload_input']
            if qfile is not None:
                content = qfile.file.read()
                form.result['upload_input.name'] = qfile.filename
                form.result['upload_input.size'] = len(content)
                form.result['upload_input.md5sum'] = hashlib.md5(content).hexdigest()

    return dict(
        breadcrumb=[(translate('Validators'), None)],
        errors=form.errors,
        result=form.result,
    )



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers application """

from pyramid.config import Configurator
from pyramid.settings import asbool

from pyramid_helpers.resources import Root


# pylint: disable=unused-argument
def main(global_config, **settings):
    """
    This function returns a Pyramid WSGI application.
    """

    # Global config
    config = Configurator(
        root_factory=Root,
        settings=settings,
    )

    # Session setup
    config.include('pyramid_beaker')

    # Mako setup
    config.include('pyramid_mako')

    # Transaction manager
    config.include('pyramid_tm')

    # Model setup
    config.include('pyramid_helpers.models')

    # Pyramid Helpers
    includeme(config)

    # Static route
    config.add_static_view('static', 'pyramid_helpers:static')

    #
    # Applicative routes
    #

    # Routes defined in views/__init__.py
    config.add_route('api-doc', '/api-doc')
    config.add_route('index', '/')
    config.add_route('i18n', '/i18n')
    config.add_route('login', '/login')
    config.add_route('logout', '/logout')
    config.add_route('predicates', '/predicates')
    config.add_route('predicates.enum', '/predicates/{predicate}', enum_predicate=dict(predicate=('value1', 'value2')))
    config.add_route('predicates.numeric-1', '/predicates/{predicate1}', numeric_predicate='predicate1')
    config.add_route('predicates.numeric-2', '/predicates/{predicate1}/{predicate2}', numeric_predicate=('predicate1', 'predicate2'))
    config.add_route('validators', '/validators')

    # Routes defined in views/articles.py
    config.add_route('articles.create', '/articles/create')
    config.add_route('articles.delete', '/articles/{article}/delete', numeric_predicate='article')
    config.add_route('articles.modify', '/articles/{article}/modify', numeric_predicate='article')
    config.add_route('articles.search', '/articles')
    config.add_route('articles.visual', '/articles/{article}', numeric_predicate='article')

    # Routes defined in views/api/articles.py
    config.add_route('api.articles.create', '/api/1.0/articles', request_method='POST')
    config.add_route('api.articles.delete', '/api/1.0/articles/{article}', numeric_predicate='article', request_method='DELETE')
    config.add_route('api.articles.modify', '/api/1.0/articles/{article}', numeric_predicate='article', request_method='PUT')
    config.add_route('api.articles.search', '/api/1.0/articles', request_method='GET')
    config.add_route('api.articles.status', '/api/1.0/articles/{article}/status', numeric_predicate='article', request_method='PUT')
    config.add_route('api.articles.visual', '/api/1.0/articles/{article}', numeric_predicate='article', request_method='GET')

    # Scan Pyramid-Helpers views modules
    config.scan('pyramid_helpers.views')

    # Create Pyramid WSGI app
    app = config.make_wsgi_app()

    return app


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers')
    """

    registry = config.registry
    settings = registry.settings

    # Utils setup
    config.include('pyramid_helpers.utils')

    # LDAP configuration
    if asbool(settings.get('ldap.enabled')):
        config.include('pyramid_helpers.ldap')

    # RADIUS configuration
    if asbool(settings.get('radius.enabled')):
        config.include('pyramid_helpers.radius')

    # Authentication setup
    if asbool(settings.get('auth.enabled')):
        config.include('pyramid_helpers.auth')

    # Forms setup
    if asbool(settings.get('forms.enabled')):
        config.include('pyramid_helpers.forms')

    # I18n setup
    if asbool(settings.get('i18n.enabled')):
        config.include('pyramid_helpers.i18n')

    # Pagers setup
    if asbool(settings.get('pagers.enabled')):
        config.include('pyramid_helpers.paginate')

    # Custom predicates
    if asbool(settings.get('predicates.enabled')):
        config.include('pyramid_helpers.predicates')

    # CSV/JSON renderers
    config.include('pyramid_helpers.renderers')



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" Forms helpers for Pyramid """

from json.decoder import JSONDecodeError
import warnings

from decorator import decorator

from formencode import Invalid
from formencode import validators
from formencode.variabledecode import variable_decode
from formencode.variabledecode import variable_encode

from pyramid.request import Request

# Monkey patch UnicodeString formencode's validator to always get unicode strings
# See: https://github.com/formencode/formencode/issues/2#issuecomment-12991249
validators.UnicodeString.outputEncoding = None

HTTP_METHODS = {
    'any': ('DELETE', 'GET', 'PATCH', 'POST', 'PUT'),
    'delete': ('DELETE', ),
    'get': ('GET', ),
    'post': ('PATCH', 'POST', 'PUT'),
}


class Form:
    """ Form validation class """

    # pylint: disable=too-many-instance-attributes

    # pylint: disable=too-many-arguments
    def __init__(self, name, schema, state, csrf_protect=False, extract='merge', method='post', persistent=False, session_key=None, volatile_items=()):
        """
        @param csrf_protect boolean, set to True is you plan to use CSRF protection
        @param extract string, where to extract data from, one of ('merge', 'get', 'post', 'json'). merge means merging GET and POST.
        @param method string, HTTP method to accept on validate(), one of ('any', 'delete', 'get', 'post'). post also includes PATCH and PUT.
        @param persistent boolean, reuse previous result if no param
        @param session_key string, custom session key for persistent storage, default is a string depending matched route name
        @param volatile_items list of string, list of items that will not be persistent
        """

        if extract not in ('merge', 'get', 'post', 'json'):
            raise ValueError(f'Invalid extract method {extract}')

        if method not in HTTP_METHODS:
            raise ValueError(f'Invalid http method {method}')

        if persistent and method != 'get':
            raise ValueError('Persistent mode is allowed only for get method')

        if session_key is None:
            # Compute unique session key for data persistence
            session_key = '[form] {0}{1}::{2}'.format(
                state.request.matched_route.name,
                '({0})'.format(
                    ', '.join(
                        '{0}={1}'.format(k, state.request.matchdict[k])
                        for k in sorted(state.request.matchdict.keys())
                    )
                ) if state.request.matchdict else '',
                name,
            )

        self.csrf_protect = csrf_protect
        self.extract = extract
        self.method = method
        self.name = name
        self.persistent = persistent
        self.schema = schema
        self.session_key = session_key
        self.state = state
        self.volatile_items = volatile_items

        self.decoded = {}
        self.encoded = {}
        self.errors = {}
        self.params = {}
        self.result = {}
        self.submitted = False
        self.valid = None

    @property
    def csrf_token(self):
        """ Get CSRF token from session """

        if not self.csrf_protect:
            return None
        request = self.state.request
        session = request.session
        return session.get_csrf_token()

    def error(self, name):
        """ Get error message for field """

        return self.errors.get(name)

    # pylint: disable=redefined-outer-name
    def from_python(self, value, validate=False):
        """
        Convert `value` from its Python representation to the foreign
        representation using formencode validator
        """

        request = self.state.request
        session = request.session

        self.decoded = self.schema.from_python(value, self.state)
        self.encoded = variable_encode(self.decoded)

        if not validate:
            return True

        try:
            self.errors = {}
            self.result = self.schema.to_python(self.decoded, self.state)
            self.valid = True

        except Invalid as exc:
            self.errors = exc.unpack_errors(variable_decode)
            self.result = {}
            self.valid = False

        if self.valid and self.persistent:
            # Store params to session
            # Restrict to form data
            session[self.session_key] = dict(
                (k, str(v))
                for k, v in self.encoded.items()
                if k not in self.volatile_items and not k.endswith('--repetitions') and v is not None
            )

            # Load params
            self.params = session[self.session_key]

            # Update request.GET so that things like pagers can work
            request.GET.update(self.params)

        return self.valid

    def to_python(self, value):
        """
        Convert `value` from its foreign representation to its Python
        representation using formencode validator
        """

        try:
            self.errors = {}
            self.result = self.schema.to_python(value, self.state)
            self.decoded = self.schema.from_python(self.result, self.state)
            self.encoded = variable_encode(self.decoded)
            self.valid = True

        except Invalid as exc:
            self.errors = exc.unpack_errors(variable_decode)
            self.result = {}
            self.decoded = value
            self.encoded = variable_encode(self.decoded)
            self.valid = False

        return self.valid

    def validate(self):
        """ Validate request parameters using formencode validator """

        request = self.state.request
        session = request.session
        translate = request.translate

        # Ensure HTTP method is the one that is expected
        if request.method not in HTTP_METHODS[self.method]:
            return None

        # Load params from request
        if self.extract == 'merge':
            self.params = request.params.mixed()

        elif self.extract == 'post':
            self.params = request.POST.mixed()

        elif self.extract == 'json':
            try:
                self.params = request.json_body
            except JSONDecodeError:
                # Invalid JSON content
                self.errors = dict(_content=translate('Invalid JSON content'))
                self.params = {}
                self.valid = False
                return False

        else:
            # get
            self.params = request.GET.mixed()

        if self.params:
            # It seems that form has been submitted
            self.submitted = True

        elif self.persistent:
            # Load params from session
            self.params = session.get(self.session_key, {})

            # Update request.GET so that things like pagers can work
            request.GET.update(self.params)

        decoded = variable_decode(self.params)

        if self.csrf_protect and self.csrf_token != decoded.get('_csrf_token'):
            # Invalid CSRF token
            self.errors = dict(_csrf_token=translate('Invalid CSRF token'))
            self.result = {}
            self.decoded = decoded
            self.encoded = variable_encode(decoded)
            self.valid = False

            return False

        # Go!
        self.to_python(decoded)

        if self.submitted and self.valid and self.persistent:
            # Store params to session
            # Restrict to form data
            session[self.session_key] = dict(
                (k, str(v))
                for k, v in self.encoded.items()
                if k not in self.volatile_items and not k.endswith('--repetitions') and v is not None
            )

        return self.valid

    def value(self, name):
        """
        Get form value for input from name
        Returned value is coerced to string for rendering purpose.

        @param string name, name of the input to get value for
        """

        if not name:
            return ''

        data = self.decoded
        for part in name.split('.'):
            index = None
            if '-' in part:
                part, index = part.split('-')
                index = int(index)

            if part not in data:
                return ''

            data = data.get(part)
            if index is not None:
                if index >= len(data):
                    return ''
                data = data[index]

        # Coerce data to string
        def coerce_str(data):
            if data is None:
                return ''

            if isinstance(data, (list, tuple)):
                return list(map(coerce_str, data))

            if isinstance(data, dict):
                return dict(
                    (k, coerce_str(v))
                    for k, v in data.items()
                )

            return str(data)

        return coerce_str(data)

    def set_data(self, value, validate=False):
        """
        Convert `value` from its Python representation to the foreign
        representation using formencode validator (deprecated)
        """

        warnings.warn('`Form.set_data()` is deprecated and will be removed soon; use `Form.from_python()`', DeprecationWarning, stacklevel=2)
        return self.from_python(value, validate=validate)


class State:
    """ Form state object """

    def __init__(self, request):
        self.request = request

    def pluralize(self, singular, plural, num, **kwargs):
        """ Wrapper to request.pluralize() """

        translated = self.request.pluralize(singular, plural, num, **kwargs)
        if translated in (singular, plural):
            # Translation failed, try the FormEncode domain
            kwargs['domain'] = 'FormEncode'
            translated = self.request.pluralize(singular, plural, num, **kwargs)

        return translated

    def translate(self, tstring, **kwargs):
        """ Wrapper to request.tramslate() """

        translated = self.request.translate(tstring, **kwargs)
        if translated == tstring:
            # Translation failed, try the FormEncode domain
            kwargs['domain'] = 'FormEncode'
            translated = self.request.translate(tstring, **kwargs)

        return translated

    # Formencode needs this to translate error messages
    _ = translate


def on_before_renderer(event):
    """ Add forms and form_ctx dictionnaries to renderer context """
    request = event['request']
    event['forms'] = request.forms
    event['form_ctx'] = request.form_ctx


def on_new_request(event):
    """
    Add forms and form_ctx dictionnaries to request

    forms dictionnary will carry all forms object using form's name as key
    form_ctx dictionnary will carry current form context:
        'current' key is reserved for current form object
        'select_options' key is reserved when building select html tag
    """
    request = event.request
    request.forms = {}
    request.form_ctx = {}


def validate(name, schema, **form_kwargs):
    """
    Form validation decorator

    Data is validated with a FormEncode schema
    """

    def wrapper(func, *args, **kwargs):
        """ Decorator Wrapper function """

        # Get request object from args
        # First arg is «self» when func is a method
        request = None
        for arg in args:
            if isinstance(arg, Request):
                request = arg
                break

        if request is None:
            raise ValueError('Missing request object in args list when calling @validate')

        # Initialize form object
        state = State(request)
        form = Form(name, schema, state, **form_kwargs)

        # Validate the form
        form.validate()

        # Store the form
        request.forms[name] = form

        return func(*args, **kwargs)

    return decorator(wrapper)


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.forms')
    """

    config.add_subscriber(on_before_renderer, 'pyramid.events.BeforeRender')
    config.add_subscriber(on_new_request, 'pyramid.events.NewRequest')



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" Database initialization script """

from argparse import ArgumentParser
import sys
import transaction

from pyramid.config import Configurator
from pyramid.paster import get_appsettings
from pyramid.paster import setup_logging

from pyramid_helpers.models import Base
from pyramid_helpers.models import DBSession
from pyramid_helpers.models.articles import Article
from pyramid_helpers.models.core import Group
from pyramid_helpers.models.core import User
from pyramid_helpers.utils import random_string


def main(argv=None):
    """ Database initialization """

    if argv is None:
        argv = sys.argv[1:]

    parser = ArgumentParser(description='Database initialization script')

    parser.add_argument(
        'config_uri',
        help='Configuration file to use.'
    )

    args = parser.parse_args(argv)

    setup_logging(args.config_uri)

    try:
        # Production
        settings = get_appsettings(args.config_uri, name='pyramid-helpers')
    except LookupError:
        # Development
        settings = get_appsettings(args.config_uri)

    config = Configurator(settings=settings)

    # Model setup
    config.include('pyramid_helpers.models')

    Base.metadata.create_all()

    # Populate table users
    with transaction.manager:
        for name in ['admin', 'guest']:
            group = DBSession.query(Group).filter(Group.name == name).first()
            if group is None:
                group = Group()
                DBSession.add(group)

            group.from_dict(dict(
                name=name,
                description=f'{name} group description',
            ))

            user = DBSession.query(User).filter(User.username == name).first()
            if user is None:
                user = User()
                DBSession.add(user)

            user.from_dict(dict(
                groups=[group],
                password=name,
                status='active',
                token=random_string(40),
                username=name,
            ))

    # Flush table articles
    with transaction.manager:
        DBSession.query(Article).delete()

    # Populate table articles
    with transaction.manager:
        author = DBSession.query(User).filter(User.username == 'admin').first()

        for i in range(1, 100):
            article = Article()
            DBSession.add(article)

            article.from_dict(dict(
                author=author,
                title=f'Article #{i}',
                text=f'Text of article #{i}',
                status='published',
            ))



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" Pagination helpers for Pyramid """

import warnings

from decorator import decorator

from pyramid.decorator import reify
from pyramid.request import Request
from pyramid.settings import aslist


LIMITS = (10, 20, 30, 50)


class Pager(list):
    """ Pagination class """

    # pylint: disable=too-many-instance-attributes

    # pylint: disable=too-many-arguments
    def __init__(self, request, name, form=None, limit=10, limits=None, sort=None, order=None, volatile_items=(), **url_kwargs):

        # Calling inherited
        super().__init__()

        self.name = name
        self.request = request
        self.defaults = dict(
            limit=limit,
            sort=sort,
            order=order,
        )

        # Keys
        self.form_key = form
        self.limit_key = '{0}.limit'.format(name)
        self.order_key = '{0}.order'.format(name)
        self.page_key = '{0}.page'.format(name)
        self.partial_key = '{0}.partial'.format(name)
        self.sort_key = '{0}.sort'.format(name)
        self.volatile_items = volatile_items + (self.page_key, self.partial_key)

        # Empty list
        self.pages = 0
        self.total = 0
        self.first_page = None
        self.last_page = None

        self.first_item = None
        self.last_item = None
        self.__page = None

        self.url_kwargs = url_kwargs

        if limits is None:
            registry = request.registry
            settings = registry.settings
            limits = settings['pagers.limits']

        self.limits = limits

    @reify
    def form(self):
        """ Associated form object """

        if self.form_key is None:
            return None

        return self.request.forms.get(self.form_key)

    @reify
    def limit(self):
        """ Limit parameter """

        limit = self.request.params.get(self.limit_key)

        try:
            limit = int(limit)
        except (TypeError, ValueError):
            limit = self.request.session.get(self.limit_key, self.defaults['limit'])

        return limit

    @reify
    def order(self):
        """ Order parameter """

        return self.request.params.get(self.order_key, self.defaults['order'])

    @property
    def page(self):
        """ Page parameter """

        if self.__page is not None:
            return self.__page

        page = self.request.params.get(self.page_key)

        try:
            page = int(page)
        except (TypeError, ValueError):
            if self.form is not None and self.form.submitted:
                # Got to first page if associated form has been submitted
                page = 1
            else:
                page = self.request.session.get(self.page_key, 1)

        self.__page = page
        return page

    @reify
    def params(self):
        """ Query parameters """

        all_params = self.request.params.mixed()

        params = dict(
            (k, v)
            for k, v in all_params.items()
            if k not in self.volatile_items
        )
        return params

    @reify
    def partial(self):
        """ Partial parameter """

        return self.partial_key in self.request.params

    @reify
    def sort(self):
        """ Sort parameter """

        return self.request.params.get(self.sort_key, self.defaults['sort'])

    def header_class(self, key):
        """ Get header class for key """

        if self.sort == key:
            extra = 'fa-sort-up' if self.order == 'desc' else 'fa-sort-down'
        else:
            extra = 'fa-sort'
        return 'fa {0}'.format(extra)

    def link(self, page=1, limit=None, sort=None, order=None):
        """ Compute a pagination link """

        params = self.params.copy()
        if page == -1:
            page = self.last_page
        params[self.page_key] = page
        if limit:
            params[self.limit_key] = limit
        if sort:
            params[self.sort_key] = sort
        if order:
            if order == 'toggle':
                if self.sort == sort:
                    order = 'asc' if self.order == 'desc' else 'desc'
                else:
                    order = 'asc'
            params[self.order_key] = order
        return self.request.current_route_path(_query=params, **self.url_kwargs)

    def links(self, before=3, after=3):
        """ Compute pagination links """

        links = []
        for page in range(self.page - before, self.page + after + 1):
            if page < self.first_page or page > self.last_page:
                continue
            links.append((page, self.link(page=page)))
        return links

    def set_collection(self, collection=None, count=None, items=None):
        """
        Set collection to pager

        @param iterable collection: A collection to get items from
        @param integer count: Collection size if collection is omitted
        @param iterable or callable items: sliced collection items or function to get items from
        """

        # Clear
        del self[:]

        if count is not None:
            self.total = count
        elif hasattr(collection, 'order_by') and hasattr(collection, 'count'):
            # SQLAlchemy: reset order by clause to speed up query
            self.total = collection.order_by(None).count()
        else:
            try:
                # Regular list
                self.total = len(collection)
            except TypeError:
                self.total = 0

        self.pages = ((self.total - 1) // self.limit) + 1
        self.first_page = 1
        self.last_page = self.pages or 1

        if self.page < self.first_page:
            self.__page = self.first_page
        elif self.page > self.last_page:
            self.__page = self.last_page

        self.first_item = (self.page - 1) * self.limit + 1
        self.last_item = min(self.first_item + self.limit - 1, self.total)

        if callable(items):
            # Get items from function
            items = items(self.first_item - 1, self.limit)
        elif collection is not None:
            # Slicing collection
            items = collection[self.first_item - 1:self.last_item]

        if items:
            self.extend(items)

        self.request.session[self.page_key] = self.page
        self.request.session[self.limit_key] = self.limit

    def to_json(self):
        """ Dump pager parameters to JSON """

        return dict(
            name=self.name,
            sort=self.sort,
            order=self.order,
            limit=self.limit,
            page=self.page,
            count=len(self),
            pages=self.pages,
            total=self.total,
        )

    @property
    def item_count(self):
        """ Getter for items count (deprecated) """
        warnings.warn('Pager.item_count` is deprecated and will be removed soon; use `Pager.total`', DeprecationWarning, stacklevel=2)
        return self.total

    @property
    def page_count(self):
        """ Getter for pages count (deprecated) """
        warnings.warn('`Pager.page_count` is deprecated and will be removed soon; use `Pager.pages`', DeprecationWarning, stacklevel=2)
        return self.pages


def on_before_renderer(event):
    """ Add pagers and pager_ctx dictionnaries to renderer context """
    request = event['request']
    event['pagers'] = request.pagers
    event['pager_ctx'] = request.pager_ctx


def on_new_request(event):
    """
    Add pagers and pager_ctx dictionnaries to request

    pagers dictionnary will carry all pagers object using pager's name as key
    pager_ctx dictionnary will carry current pager context:
        'current' key is reserved for current pager object
    """
    request = event.request
    request.pagers = {}
    request.pager_ctx = {}


def paginate(name, partial_template=None, **pager_kwargs):
    """ Pagination decorator """

    def wrapper(func, *args, **kwargs):
        """ Decorator Wrapper function """

        # Get request object from args
        # First arg is «self» when func is a method
        request = None
        for arg in args:
            if isinstance(arg, Request):
                request = arg
                break

        if request is None:
            raise ValueError('Missing request object in args list when calling @paginate')

        # Initialize pager object
        pager = Pager(request, name, **pager_kwargs)

        # Store the pager
        request.pagers[name] = pager

        # Call function
        result = func(*args, **kwargs)

        if not request.is_response(result) and pager.partial and partial_template:
            # Override template if needed
            request.override_renderer = partial_template

        return result

    return decorator(wrapper)


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.paginate')
    """

    registry = config.registry
    settings = registry.settings

    try:
        limits = list(map(int, aslist(settings['pagers.limits'])))
        # Add default value if missing
        if 10 not in limits:
            limits.append(10)
        limits.sort()
    except (KeyError, ValueError):
        limits = LIMITS

    settings['pagers.limits'] = limits

    config.add_subscriber(on_before_renderer, 'pyramid.events.BeforeRender')
    config.add_subscriber(on_new_request, 'pyramid.events.NewRequest')



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Authentication forms """

import formencode
from formencode import validators


# login
class LoginForm(formencode.Schema):
    """
    :param username: The username to authenticate
    :param password: The password to validate
    :param redirect: URL to redirect to if authentication is successful
    """

    allow_extra_fields = True
    filter_extra_fields = True

    username = validators.String(not_empty=True)
    password = validators.String(not_empty=True)
    redirect = validators.String(if_missing=None, if_empty=None)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers views for articles management """

from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPNotFound
from pyramid.view import view_config

from pyramid_helpers.forms import validate
from pyramid_helpers.forms.articles import CreateForm
from pyramid_helpers.funcs.articles import create_or_modify
from pyramid_helpers.funcs.articles import build_search_query
from pyramid_helpers.funcs.articles import get_article
from pyramid_helpers.models import DBSession
from pyramid_helpers.paginate import paginate


@view_config(permission='articles.create', route_name='articles.create', renderer='/articles/modify.mako')
@validate('article', CreateForm)
def create(request):
    """ Creates an article """

    translate = request.translate
    form = request.forms['article']

    if request.method == 'POST':
        article = create_or_modify(request, form)
        if article:
            return HTTPFound(location=request.route_path('articles.visual', article=article.id))

    breadcrumb = [
        (translate('Articles'), request.route_path('articles.search')),
        (translate('New article'), request.route_path('articles.create'))
    ]
    title = translate('New article')
    cancel_link = request.route_path('articles.search')

    return dict(
        breadcrumb=breadcrumb,
        cancel_link=cancel_link,
        title=title,
    )


@view_config(permission='articles.delete', route_name='articles.delete', renderer='/confirm.mako')
def delete(request):
    """ Deletes an article """

    translate = request.translate

    article = get_article(request)
    if article is None:
        raise HTTPNotFound(detail=translate('Invalid article'))

    if request.method == 'POST':
        if 'cancel' in request.params:
            return HTTPFound(location=request.route_url('articles.visual', article=article.id))

        DBSession.delete(article)
        return HTTPFound(location=request.route_url('articles.search'))

    breadcrumb = [
        (translate('Articles'), request.route_path('articles.search')),
        (article.title, request.route_path('articles.visual', article=article.id)),
        (translate('Deletion'), request.route_path('articles.delete', article=article.id)),
    ]
    return dict(
        breadcrumb=breadcrumb,
        note=None,
        question=translate('Do you really want to delete article "{0}" ?').format(article.title),
        title=translate('Article "{0}" deletion').format(article.title),
    )


@view_config(permission='articles.modify', route_name='articles.modify', renderer='/articles/modify.mako')
@validate('article', CreateForm)
def modify(request):
    """ Modifies an article """

    translate = request.translate
    form = request.forms['article']

    article = get_article(request)
    if article is None:
        raise HTTPNotFound(detail=translate('Invalid article'))

    if request.method == 'POST':
        if create_or_modify(request, form, article=article):
            return HTTPFound(location=request.route_url('articles.visual', article=article.id))
    else:
        data = article.to_dict()
        form.from_python(data)

    breadcrumb = [
        (translate('Articles'), request.route_path('articles.search')),
        (article.title, request.route_path('articles.visual', article=article.id)),
        (translate('Edition'), request.route_path('articles.modify', article=article.id)),
    ]
    title = translate('Article "{0}" edition').format(article.title)
    cancel_link = request.route_url('articles.visual', article=article.id)

    return dict(
        breadcrumb=breadcrumb,
        cancel_link=cancel_link,
        title=title,
    )


@view_config(permission='articles.visual', route_name='articles.search', renderer='/articles/search.mako')
@paginate('articles', limit=10, sort='id', order='desc', partial_template='/articles/list.mako')
def search(request):
    """ Search for articles """

    translate = request.translate
    pager = request.pagers['articles']

    if not request.has_permission('articles.modify'):
        criteria = dict(status='published')
    else:
        criteria = {}

    query = build_search_query(order=pager.order, sort=pager.sort, **criteria)
    pager.set_collection(query)

    if pager.partial:
        return dict()

    if 'csv' in request.params:
        # CSV export
        request.override_renderer = 'csv'

        rows = []

        # Add header
        rows.append([
            translate('id'),
            translate('title'),
            translate('creation date'),
            translate('status'),
            translate('text'),
        ])

        # Add content
        rows.extend([
            [article.id, article.title, article.creation_date, article.status, article.text]
            for article in query
        ])

        return dict(
            delimiter=';',
            filename=translate('articles.csv'),
            rows=rows,
        )

    breadcrumb = [
        (translate('Articles'), request.route_path('articles.search'))
    ]

    return dict(
        breadcrumb=breadcrumb,
        title=translate('Articles')
    )


@view_config(permission='articles.visual', route_name='articles.visual', renderer='/articles/visual.mako')
def visual(request):
    """ Visualizes an article """

    translate = request.translate

    article = get_article(request)
    if article is None:
        raise HTTPNotFound(detail=translate('Invalid article'))

    breadcrumb = [
        (translate('Articles'), request.route_path('articles.search')),
        (article.title, request.route_path('articles.visual', article=article.id)),
    ]
    return dict(
        article=article,
        breadcrumb=breadcrumb,
        title=article.title
    )



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Custom renderers for Pyramid-Helpers """

import csv
import io
import json
from logging import getLogger

from pyramid.httpexceptions import HTTPInternalServerError
from pyramid.settings import asbool


log = getLogger(__name__)


# pylint: disable=unused-argument
def csv_renderer_factory(info):
    """ CSV renderer for Pyramid """

    def _render(value, system):
        request = system.get('request')
        translate = request.translate

        # Get parameters
        encoding = value.pop('encoding', 'utf-8')
        filename = value.pop('filename', 'result.csv')
        rows = value.pop('rows', None) or []

        # Create CSV
        try:
            fp = io.StringIO()

            writer = csv.writer(fp, **value)
            writer.writerows(rows)

        except Exception as exc:
            log.exception('Could not convert view response to CSV')
            raise HTTPInternalServerError(detail=translate('Could not convert view response to CSV')) from exc

        # Set content type
        request.response.content_type = 'text/csv; charset="{0}"'.format(encoding)
        request.response.content_disposition = 'attachment; filename="{0}"'.format(filename)

        # Return file content
        fp.seek(0)
        content = fp.read()
        return content.encode(encoding)

    return _render


# pylint: disable=unused-argument
def json_renderer_factory(info):
    """ Custom JSON renderer with callback suppport """

    def _render(value, system):
        request = system.get('request')
        registry = request.registry
        settings = registry.settings

        # Prepare options
        kwargs = dict(
            (k[15:], v)
            for k, v in settings.items()
            if k.startswith('renderers.json.')
        )

        kwargs.pop('enabled')

        if 'indent' in kwargs:
            kwargs['indent'] = int(kwargs['indent'])

        for key in ('skipkeys', 'ensure_ascii', 'check_circular', 'allow_nan', 'sort_keys'):
            if key in kwargs:
                kwargs[key] = asbool(kwargs[key])

        # Extrack callback from query
        callback = kwargs.pop('callback', None)
        if callback:
            callback = request.params.get(callback)

        result = json.dumps(value, default=str, **kwargs)

        if callback:
            request.response.content_type = 'application/javascript; charset="utf-8"'
            result = '{0}({1})'.format(callback, result)
        else:
            request.response.content_type = 'application/json; charset="utf-8"'

        return result

    return _render


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.renderers')
    """

    registry = config.registry
    settings = registry.settings

    if asbool(settings.get('renderers.csv.enabled')):
        # Add CSV renderer
        config.add_renderer('csv', csv_renderer_factory)

    if asbool(settings.get('renderers.json.enabled')):
        # Replace JSON renderer (callback support)
        config.add_renderer('json', json_renderer_factory)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers resources """

from pyramid.authorization import ALL_PERMISSIONS
from pyramid.authorization import Allow
from pyramid.authorization import Everyone


class Root:
    """ Pyramid-Helpers resources """

    # pylint: disable=too-few-public-methods

    __acl__ = [
        (Allow, Everyone, (
            'articles.search',
            'articles.visual',
        )),

        (Allow, 'group:admin', ALL_PERMISSIONS),

        (Allow, 'group:guest', (
            'articles.create',
            'articles.modify',
        )),
    ]

    def __init__(self, request):
        self.request = request



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers core models """

import datetime
import logging

from passlib.context import CryptContext

from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column
from sqlalchemy.schema import Table
from sqlalchemy.schema import ForeignKey
from sqlalchemy.types import DateTime
from sqlalchemy.types import Enum
from sqlalchemy.types import Integer
from sqlalchemy.types import Text
from sqlalchemy.types import Unicode

from pyramid_helpers.auth import AuthenticationClient
from pyramid_helpers.models import Base
from pyramid_helpers.models import DBSession


CRYPT_SCHEMES = ['argon2', 'bcrypt', 'pbkdf2_sha256', 'pbkdf2_sha512', 'sha256_crypt', 'sha512_crypt']


log = logging.getLogger(__name__)


# Users <-> Groups
user_groups_associations = Table(
    'user_groups_associations',
    Base.metadata,
    Column('user_id', Integer, ForeignKey('users.id')),
    Column('group_id', Integer, ForeignKey('groups.id')),
)


class DatabaseAuthenticationClient(AuthenticationClient):
    """ Authentication client for database """

    # pylint: disable=too-few-public-methods

    __auth_mode__ = 'database'

    # pylint: disable=no-self-use,unused-argument
    def validate_password(self, username, password, request):
        """ Validate password """

        user = get_user_by_username(username, request)
        if user is None:
            return False

        return user.validate_password(password)


class Group(Base):
    """ ORM class mapped to groups table """

    __tablename__ = 'groups'

    # Primary key
    id = Column(Integer, primary_key=True)

    # Attributes
    creation_date = Column(DateTime, nullable=False)
    description = Column(Text)
    modification_date = Column(DateTime, nullable=False)
    name = Column(Unicode(255), unique=True)

    def from_dict(self, data):
        """ Load data from dict """

        utcnow = datetime.datetime.utcnow()

        if 'description' in data:
            self.description = data['description']

        if 'name' in data:
            self.name = data['name']

        if self.creation_date is None:
            self.creation_date = data.get('creation_date') or utcnow

        if data.get('modification_date'):
            self.modification_date = data['modification_date']

        elif DBSession.is_modified(self):
            self.modification_date = utcnow

        return DBSession.is_modified(self)

    def to_dict(self, context=None):
        """ Dump data to dict """

        data = dict(
            name=self.name,
            description=self.description,
        )

        if context == 'api.search':
            data.update(dict(
                creation_date=self.creation_date,
                modification_date=self.modification_date,
            ))
            return data

        return data

    def to_json(self, context=None):
        """ Dump data to JSON """

        return self.to_dict(context=context)


class User(Base):
    """ ORM class mapped to users table """

    # pylint: disable=too-many-instance-attributes

    __tablename__ = 'users'

    # Primary key
    id = Column(Integer, primary_key=True)

    # Attributes
    creation_date = Column(DateTime, nullable=False)
    firstname = Column(Unicode(255))
    lastname = Column(Unicode(255))
    modification_date = Column(DateTime, nullable=False)
    password = Column(Unicode(32))
    status = Column(Enum('active', 'disabled'))
    timezone = Column(Unicode(255))
    token = Column(Unicode(255))
    username = Column(Unicode(255), unique=True, nullable=False)

    # Relations
    groups = relationship('Group', secondary=user_groups_associations)

    @property
    def fullname(self):
        """ Compute fullname from firstname, lastname or username """

        fullname = []
        if self.firstname:
            fullname.append(self.firstname)
        if self.lastname:
            fullname.append(self.lastname)
        if not fullname:
            fullname.append(self.username)
        return ' '.join(fullname)

    def from_dict(self, data):
        """ Load data from dict """

        utcnow = datetime.datetime.utcnow()

        if 'firstname' in data:
            self.firstname = data['firstname']

        if 'lastname' in data:
            self.lastname = data['lastname']

        if 'password' in data and not self.validate_password(data['password']):
            self.set_password(data['password'])

        if 'groups' in data:
            self.groups = data['groups']

        if 'status' in data:
            self.status = data['status']

        if 'timezone' in data:
            self.timezone = data['timezone']

        if 'token' in data:
            self.token = data['token']

        if 'username' in data:
            self.username = data['username']

        if self.creation_date is None:
            self.creation_date = data.get('creation_date') or utcnow

        if data.get('modification_date'):
            self.modification_date = data['modification_date']

        elif DBSession.is_modified(self):
            self.modification_date = utcnow

        return DBSession.is_modified(self)

    def to_json(self, context=None):
        """ Dump data to JSON """

        data = dict(
            id=self.id,
            firstname=self.firstname,
            lastname=self.lastname,
        )

        if context == 'api.search':
            data.update(dict(
                creation_date=self.creation_date,
                modification_date=self.modification_date,
            ))
            return data

        data.update(dict(
            username=self.username,
            groups=[group.to_json() for group in self.groups]
        ))

        return data

    def set_password(self, password, scheme='sha512_crypt'):
        """ Hash and set password """

        if password is None:
            self.password = None
            return

        ctx = CryptContext(default=scheme, schemes=CRYPT_SCHEMES)
        self.password = ctx.hash(password)

    def validate_password(self, password):
        """ Validate password """

        if self.password is None:
            return False

        if self.status != 'active':
            return False

        ctx = CryptContext(schemes=CRYPT_SCHEMES)
        try:
            validated = ctx.verify(password, self.password)
        except ValueError:
            log.exception('Failed to verify password using CryptContext.verify() for user #%s', self.id)
            validated = False

        return validated


# pylint: disable=unused-argument
def get_principals(username, request):
    """ Get principals for user """

    user = get_user_by_username(username, request)
    if user is None:
        # Invalid user
        return None

    return [
        ('group:{0}'.format(group.name))
        for group in user.groups
    ]


# pylint: disable=unused-argument
def get_user_by_username(username, request):
    """ Get user from database by username """

    if username is None:
        return None

    user = DBSession.query(User).filter_by(username=username).first()
    if user is None:
        log.warning('Failed to get user with username=%s', username)

    return user


# pylint: disable=unused-argument
def get_username_by_token(token, request):
    """ Get user from database by token """

    if token is None:
        return None

    user = DBSession.query(User).filter_by(token=token).first()
    if user is None:
        log.warning('Failed to get user with token=%s', token)
        return None

    return user.username



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Custom predicates for Pyramid """


class EnumPredicate:
    """
    The route defined by:

    config.add_route('route-name', '/foo/{kind}/bar', enum_predicate=dict(kind=('one', 'two')))

    Will only match the following urls:
     * /foo/one/bar
     * /foo/two/bar
    """

    # pylint: disable=unused-argument
    def __init__(self, params, config):
        self.names = list(params)
        self.params = params

    def __call__(self, context, request):
        matchdict = context['match']
        for name, values in self.params.items():
            if matchdict.get(name) not in values:
                return False
        return True

    def text(self):
        """ Predicate identifier """
        return '{0}'.format(self.params)

    phash = text


class NumericPredicate:
    """
    The route defined by:

    config.add_route('route-name', '/foo/{id}/bar', numeric_predicate='id')

    Will only match the following urls:
     * /foo/[0-9]+/bar
    """

    # pylint: disable=unused-argument
    def __init__(self, names, config):
        if isinstance(names, str):
            self.names = (names, )
        else:
            self.names = names

    def __call__(self, context, request):
        matchdict = context['match']
        for name in self.names:
            if not matchdict.get(name).isnumeric():
                return False
        return True

    def text(self):
        """ Predicate identifier """
        return '{0}'.format(self.names)

    phash = text


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.predicates')
    """

    config.add_route_predicate('enum_predicate', EnumPredicate)
    config.add_route_predicate('numeric_predicate', NumericPredicate)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers data models """

import logging

from sqlalchemy import engine_from_config
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker

from zope.sqlalchemy import register


DBSession = scoped_session(sessionmaker())
register(DBSession)
Base = declarative_base()

log = logging.getLogger(__name__)


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.models')
    """

    registry = config.registry
    settings = registry.settings

    # Initialize model
    engine = engine_from_config(settings, 'sqlalchemy.')
    DBSession.configure(bind=engine)

    # pylint: disable=import-outside-toplevel
    from pyramid_helpers.models.core import DatabaseAuthenticationClient

    # Register authentication class
    auth_client = DatabaseAuthenticationClient()
    auth_client.setup()

    log.info('[MODELS] Initialization complete: dialect=%s, engine=%s, url=%r', engine.dialect.name, engine.driver, engine.url)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Articles forms """

import formencode
from formencode import validators

from .validators import List


# articles.create, articles.modify
# api.articles.create, api.articles.modify
class CreateForm(formencode.Schema):
    """
    :param title: Title
    :param text: Text
    :param status: Status
    """
    allow_extra_fields = True
    filter_extra_fields = True

    title = validators.String(not_empty=True)
    text = validators.String(not_empty=True)
    status = validators.OneOf(['draft', 'published', 'refused'], not_empty=True)


# api.articles.status
class StatusForm(formencode.Schema):
    """
    :param status: Status
    """
    allow_extra_fields = True
    filter_extra_fields = True

    status = validators.OneOf(['draft', 'published', 'refused'], not_empty=True)


# api.articles.search
class SearchForm(formencode.Schema):
    """
    :param exact: Text search strings must exact match the title of article
    :param excluded_ids: List of article Ids to exclude from result
    :param selected_ids: List of article Ids to include in result
    :param status: Status of article
    :param term: Text search string used to filter articles. Only articles containing the term in their title will be returned
    :param text: Text
    :param title: Title
    """
    allow_extra_fields = True
    filter_extra_fields = True

    exact = validators.StringBool(if_missing=False, if_empty=False)
    excluded_ids = List(validators.Int())
    selected_ids = List(validators.Int())
    status = validators.OneOf(['draft', 'published', 'refused'], if_empty=None, if_missing=None)
    term = validators.String(if_missing=None, if_empty=None)
    text = validators.String(if_empty=None, if_missing=None)
    title = validators.String(if_empty=None, if_missing=None)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers functions for articles management """

import sqlalchemy as sa

from pyramid_helpers.models import DBSession
from pyramid_helpers.models.articles import Article


def build_search_query(sort='id', order='asc', **criteria):
    """ Build a search query from criteria """

    query = DBSession.query(Article)

    # Filters
    if criteria.get('excluded_ids'):
        query = query.filter(~Article.id.in_(criteria['excluded_ids']))

    if criteria.get('selected_ids'):
        query = query.filter(Article.id.in_(criteria['selected_ids']))

    if criteria.get('term'):
        if criteria.get('exact') is True:
            query = query.filter(Article.title == criteria['term'])
        else:
            query = query.filter(
                Article.title.ilike('%{0}%'.format(criteria['term'].replace(' ', '%'))),
            )

    if criteria.get('title'):
        query = query.filter(
            Article.title.ilike('%{0}%'.format(criteria['title'].replace(' ', '%'))),
        )

    if criteria.get('text'):
        query = query.filter(
            Article.text.ilike('%{0}%'.format(criteria['text'].replace(' ', '%'))),
        )

    if criteria.get('status'):
        query = query.filter(Article.status == criteria['status'])

    # Order
    if sort == 'id':
        order_by = [Article.id]
    elif sort == 'creation_date':
        order_by = [Article.creation_date]
    elif sort == 'modification_date':
        order_by = [Article.modification_date]
    elif sort == 'title':
        order_by = [Article.title]
    elif sort == 'status':
        order_by = [Article.status]

    sa_func = sa.asc if order == 'asc' else sa.desc
    query = query.order_by(*[sa_func(column) for column in order_by])

    return query


def create_or_modify(request, form, article=None):
    """ Create or modify function """

    translate = request.translate

    if form.errors:
        return None

    # Check title unicity
    same_title = DBSession.query(Article).filter_by(title=form.result['title']).first()
    if same_title and same_title != article:
        form.errors['title'] = translate('Title already used in another article')
        return None

    # Do the job
    if article is None:
        article = Article()
        DBSession.add(article)

    # Set author
    form.result['author'] = request.authenticated_user

    # Update object
    article.from_dict(form.result)

    # Flushing the session to get new id
    DBSession.flush()

    return article


def get_article(request):
    """ Extract article id from query and return corresponding database object """

    article_id = request.matchdict.get('article')
    if article_id is None:
        return None

    return DBSession.get(Article, article_id)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Radius client for Pyramid-Helpers """

from io import StringIO
import logging
import threading
import time

from beaker.util import NoneType

from pyrad.client import Client
from pyrad.client import Timeout
from pyrad.dictionary import Dictionary
from pyrad.packet import AccessAccept
from pyrad.packet import AccessRequest

from pyramid_helpers.auth import AuthenticationClient
from pyramid_helpers.utils import ConfigurationError
from pyramid_helpers.utils import get_settings
from pyramid_helpers.utils import parse_settings


SETTINGS_DEFAULTS = {
    'radius': {
        'enabled': False,
        'retries': 2,
        'timeout': 3.0,
    },
    r'server:\d+': {
        'enabled': True,
        'acctport': 1813,
        'authport': 1812,
        'coaport': 3799,
    },
}

SETTINGS_RULES = {
    r'attr:\d+': [
        ('name', (str, ), 'name must be a string designating valid attribute name'),
        ('type', (str, ), 'type must be a string designating valid attribute type'),
    ],
    'radius': [
        ('enabled', (bool, NoneType), 'enabled must be a boolean or an integer'),
        ('retries', (int, ), 'retries must be a string designating a valid integer'),
        ('timeout', (float, ), 'timeout must be a string designating a valid float'),
    ],
    r'server:\d+': [
        ('enabled', (bool, NoneType), 'enabled must be a boolean or an integer'),
        ('server', (str, ), 'server must be a string designating valid server'),
        ('secret', (str, ), 'secret must be a string designating a valid secret'),
        ('authport', (int, ), 'authentication port must be an integer'),
        ('acctport', (int, ), 'accounting port must be an integer'),
        ('coaport', (int, ), 'CoA port must be an integer'),
    ],
}


log = logging.getLogger(__name__)

client = None


class RadiusError(Exception):
    """ RADIUS error """


class RadiusClient(AuthenticationClient):
    """ RADIUS authentication client """

    __auth_mode__ = 'radius'

    def __init__(self):
        """ RADIUS initialization """

        self.dictionary = None
        self.retries = None
        self.servers = []
        self.timeout = None

        self.__lock = threading.Lock()

    def __enter__(self):
        """ Acquire lock """

        # pylint: disable=consider-using-with
        self.__lock.acquire()

        return self

    def __exit__(self, type_, value, traceback):
        """ Release lock """

        self.__lock.release()

    def setup(self, *args, **kwargs):
        """
        RADIUS client setup

        :param servers: List or server settings to connect to
        :param attributes: Attributes to use
        :param retries: Integer, maximum retries before stopping to query the server
        :param timeout: Float, timeout value for connections
        """

        # Prepare dictionary
        self.dictionary = []

        attributes = kwargs.pop('attributes', None) or []
        for data in attributes:
            flags = data.get('flags')
            self.dictionary.append('ATTRIBUTE {name:20s} {oid:5d} {type:10s}{0}'.format(' {0}'.format(flags) if flags else '', **data))

        # Prepare clients settings
        self.servers = kwargs.pop('servers', None) or []

        if kwargs.get('retries'):
            self.retries = kwargs.pop('retries')

        if kwargs.get('timeout'):
            self.timeout = kwargs.pop('timeout')

        # Calling inherited
        return super().setup(*args, **kwargs)

    def validate_password(self, username, password, request):
        """ Attempt to send an authentication packet to servers """

        if not self.__lock.locked():
            raise RadiusError('Please use RadiusClient inside a `with` statement')

        now = time.time()

        for settings in sorted(self.servers, key=lambda settings: settings.get('date', now)):
            kwargs = settings.copy()
            kwargs.pop('date', None)

            kwargs['dict'] = Dictionary(StringIO('\n'.join(self.dictionary)))

            client_ = Client(**kwargs)

            if self.retries:
                client_.retries = self.retries

            if self.timeout:
                client_.timeout = self.timeout

            req = client_.CreateAuthPacket(code=AccessRequest, username=username)
            req['password'] = req.PwCrypt(password)

            try:
                reply = client_.SendPacket(req)
            except Timeout:
                log.exception('Failed to communicate with RADIUS server (server=%s)', client_.server)
                continue

            client_._CloseSocket()    # pylint: disable=protected-access

            settings['date'] = now

            if reply.code == AccessAccept:
                return True

            # Invalid user
            log.error('[RADIUS] Authentication failed (server=%s, username=%s)', client_.server, username)

        return False


# Create client object
client = RadiusClient()


def includeme(config):
    """ RADIUS module initialization """

    # Load and parse settings
    settings = get_settings(config, 'radius')
    if settings is None:
        raise ConfigurationError('[RADIUS] Invalid or missing configuration for RADIUS, please check radius.filepath directive')

    settings = parse_settings(settings, SETTINGS_RULES, defaults=SETTINGS_DEFAULTS)

    # Settings
    attributes = []
    servers = []

    for section in settings:
        if section.startswith('attr:'):
            data = settings[section].copy()
            try:
                data['oid'] = int(section[5:])
            except ValueError as exc:
                raise ConfigurationError(f'[RADIUS] Invalid attribute definition for {section} in configuration') from exc

            attributes.append(data)

        elif section.startswith('server:'):
            data = settings[section].copy()
            if not data.pop('enabled'):
                continue

            data['secret'] = bytes(data['secret'], 'utf-8')

            servers.append(data)

    if not servers:
        raise ConfigurationError('[RADIUS] Empty or invalid configuration for server entries in configuration')

    # Prepare arguments
    kwargs = settings['radius'].copy()
    kwargs['attributes'] = attributes
    kwargs['servers'] = servers

    # Setup the client
    client.setup(**kwargs)

    log.info('[RADIUS] Initialization complete: servers=%s', ','.join(kwargs['server'] for kwargs in client.servers))



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" API REST interactive documentation view """

import inspect
import logging
import re

from formencode.api import NoDefault

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.static import static_view


log = logging.getLogger(__name__)

DECORATOR_RE = re.compile(r'^@(?P<func>paginate|validate)\((?P<args>.*)?\)$')

PAGER_PARAMS = {
    'sort': dict(description='Sorting key', type='UnicodeString'),
    'order': dict(description='Sorting order', type='OneOf', values=['asc', 'desc']),
    'limit': dict(description='Number of items per page', type='OneOf'),
    'page': dict(description='Page number', type='Int', default=1),
}


def datetime_format_python_to_moment(format_):
    """ Convert python datetime format (C Standard 1989) into Moment.js datetime format """

    # year 4 digits
    format_ = format_.replace('%Y', 'YYYY')
    # month 2 digits
    format_ = format_.replace('%m', 'MM')
    # day 2 digits
    format_ = format_.replace('%d', 'DD')
    # hours 2 digits (24h)
    format_ = format_.replace('%H', 'HH')
    # hours 2 digits (12h)
    format_ = format_.replace('%I', 'hh')
    # hours AM/PM
    format_ = format_.replace('%p', 'A')
    # minutes 2 digits
    format_ = format_.replace('%M', 'mm')
    # seconds 2 digits
    format_ = format_.replace('%S', 'ss')

    return format_


def inspect_schema_doc(tree, parameters, parent=None):
    """ Inspect `formencode.Schema` documentation """

    if not parameters:
        return

    if isinstance(tree, list):
        for node in tree:
            inspect_schema_doc(node, parameters)
        return

    if tree[0].__name__ in ('object', 'Declarative') or tree[0].__module__.startswith('formencode'):
        return

    doc = inspect.getdoc(tree[0])
    for line in doc.split('\n'):
        if not line.startswith(':param'):
            continue

        line = line.split(':', 2)
        if len(line) != 3:
            continue

        name = line[1].split(' ')[-1].strip()
        if parent is not None:
            name = '{0}.{1}'.format(parent, name)
        if name in parameters:
            parameters[name]['description'] = line[2].strip()


def inspect_validator(validator, name=None):
    """ Inspect `formencode.validators.FancyValidator` """

    data = dict(
        description=None,
    )

    if name is not None:
        data['name'] = name
        data['required'] = validator.not_empty is True
        data['parameter_type'] = 'query'

    if_empty = None
    if validator.if_empty is not None and validator.if_empty is not NoDefault:
        if_empty = validator.if_empty
    if_missing = None
    if validator.if_missing is not None and validator.if_missing is not NoDefault:
        if_missing = validator.if_missing
    data['default'] = if_missing or if_empty

    class_name = validator.__class__.__name__
    if class_name.endswith('Validator'):
        class_name = class_name.replace('Validator', '')

    data['type'] = class_name
    if class_name == 'OneOf':
        data['values'] = validator.list
    elif class_name in ('ForEach', 'NumberList', 'StringList'):
        data['items'] = inspect_validator(validator.validators[0]) if validator.validators else None
    elif class_name in ('Int', 'Number'):
        data['min'] = validator.min
        data['max'] = validator.max
    elif data['type'] == 'DateTime':
        if validator.is_date:
            data['type'] = 'Date'
        data['format'] = datetime_format_python_to_moment(validator.format)

    return data


# from CPython project
# https://hg.python.org/cpython/file/tip/Lib/inspect.py
def unwrap(func, stop=None):
    """Get the object wrapped by *func*.

    Follows the chain of :attr:`__wrapped__` attributes returning the last
    object in the chain.

    *stop* is an optional callback accepting an object in the wrapper chain
    as its sole argument that allows the unwrapping to be terminated early if
    the callback returns a true value. If the callback never returns a true
    value, the last object in the chain is returned as usual. For example,
    :func:`signature` uses this to stop unwrapping if any object in the
    chain has a ``__signature__`` attribute defined.

   :exc:`ValueError` is raised if a cycle is encountered.

    """
    if stop is None:
        def _is_wrapper(func):
            return hasattr(func, '__wrapped__')
    else:
        def _is_wrapper(func):
            return hasattr(func, '__wrapped__') and not stop(func)

    # remember the original func for error reporting
    func_ = func

    # Memoise by id to tolerate non-hashable objects
    memo = set([id(func_)])
    while _is_wrapper(func):
        func = func.__wrapped__
        id_func = id(func)
        if id_func in memo:
            raise ValueError('wrapper loop when unwrapping {!r}'.format(func_))
        memo.add(id_func)

    return func


# pylint: disable=too-many-branches,too-many-locals,too-many-statements
def api_doc(request, renderer_values):
    """Render API documentation

    Args:
        request (obj): currently active request.
        renderer_values (dict): values passed to renderer.
            breadcrumb: Page breadcrumb
            hide_undocumented: True to hide undocumented services (default False)
            title: Page title
            subtitle: Page subtitle

    Returns:
        unicode: HTML response body.

    """

    registry = request.registry
    settings = registry.settings
    translate = request.translate

    introspector = registry.introspector

    # Get defined routes
    services = dict()

    for item in introspector.get_category('routes'):
        introspectable = item['introspectable']
        route_name = introspectable.discriminator

        if not route_name.startswith('api.'):
            continue

        route_intr = introspector.get('routes', route_name)
        route = route_intr['object']

        services[route_name] = dict(
            allowed=True,
            doc=None,
            module=None,
            name=route_name,
            pattern=request.script_name + route_intr['pattern'],
            parameters=[],
            request_methods=route_intr['request_methods'],
        )

        # get custom predicates
        custom_predicates = []
        for predicate in route.predicates:
            if predicate.__class__.__name__ == 'EnumPredicate':
                param_type = 'OneOf'
            elif predicate.__class__.__name__ == 'NumericPredicate':
                param_type = 'Int'
            else:
                continue

            for name in predicate.names:
                param_data = dict(
                    name=name,
                    type=param_type,
                    parameter_type='path',
                    required=True,
                    default=None,
                    description=None,
                )

                if predicate.__class__.__name__ == 'EnumPredicate':
                    param_data['values'] = predicate.params[name]

                services[route_name]['parameters'].append(param_data)

                custom_predicates.append(name)

        # get other predicates (not custom) present in pattern
        for name in route.match(route_intr['pattern']).keys():
            if name not in custom_predicates:
                services[route_name]['parameters'].append(dict(
                    name=name,
                    type='ByteString',
                    parameter_type='path',
                    required=True,
                    default=None,
                    description=None,
                ))

    # Check permission
    for item in introspector.get_category('permissions'):
        introspectable = item['introspectable']
        permission = introspectable.discriminator
        if permission == NO_PERMISSION_REQUIRED:
            continue

        for introspectable_view in item['related']:
            route_name = introspectable_view['route_name']
            if route_name in services:
                services[route_name]['allowed'] = bool(request.has_permission(permission))

    # Get associated views
    modules = dict()

    for item in introspector.get_category('views'):
        introspectable = item['introspectable']
        route_name = introspectable['route_name']
        view = introspectable['callable']

        service = services.get(route_name)
        if service is None:
            continue

        if introspectable['request_methods']:
            if isinstance(introspectable['request_methods'], tuple):
                service['request_methods'] = introspectable['request_methods']
            else:
                service['request_methods'] = (introspectable['request_methods'],)

        callable_func = unwrap(view)

        __globals__ = callable_func.__globals__
        module_name = __globals__['__name__']

        if module_name not in modules:
            modules[module_name] = dict(
                doc=callable_func.__globals__['__doc__'].strip(),
                services=[],
            )
        modules[module_name]['services'].append(service)

        service['module'] = module_name
        if callable_func.__doc__:
            service['doc'] = callable_func.__doc__.strip()

        source_lines, _ = inspect.getsourcelines(callable_func)
        for line in source_lines:
            if line.startswith('def '):
                break

            # Extract decorator parameters
            match = DECORATOR_RE.match(line)
            if match is None:
                continue

            decorator_name, params_str = match.groups()

            def __extract__(*args, **kwargs):
                __extract__.args = args
                __extract__.kwargs = kwargs

            globals_ = __globals__.copy()
            globals_['__extract__'] = __extract__

            # pylint: disable=eval-used
            eval(f'__extract__({params_str})', globals_)

            if decorator_name == 'paginate':
                pager_name = __extract__.args[0]

                for param_name, param_data in PAGER_PARAMS.items():
                    param_data = param_data.copy()

                    param_data.update(dict(
                        name=f'{pager_name}.{param_name}',
                        required=False,
                        parameter_type='query (paging)',
                    ))

                    if param_name in __extract__.kwargs:
                        param_data['default'] = __extract__.kwargs[param_name]
                    elif 'default' not in param_data:
                        param_data['default'] = None

                    if param_name == 'limit':
                        param_data['values'] = __extract__.kwargs.get('limits', settings['pagers.limits'])

                    service['parameters'].append(param_data)

            elif decorator_name == 'validate':
                schema = __extract__.args[1]

                parameters = dict()
                for name, field in schema.fields.items():
                    class_name = field.__class__.__name__
                    if class_name.endswith('Schema'):
                        for sub_name, sub_field in field.__dict__['fields'].items():
                            parameter_name = '{0}.{1}'.format(name, sub_name)
                            parameters[parameter_name] = inspect_validator(sub_field, parameter_name)

                        inspect_schema_doc((field.__class__,), parameters, name)
                    else:
                        parameters[name] = inspect_validator(field, name)

                inspect_schema_doc(inspect.getclasstree(inspect.getmro(schema), unique=True), parameters)

                # sort parameters by type and name
                service['parameters'] += sorted(parameters.values(), key=lambda param_data: (param_data['type'], param_data['name']))

    # Orphan routes
    orphans = [
        service
        for service in services.values()
        if service['module'] is None
    ]

    # Get Boostrap version
    bootstrap_version = settings.get('api_doc.bootstrap_version', None)
    if bootstrap_version is None:
        raise HTTPBadRequest(detail=translate('Missing `api_doc.bootstrap_version` entry in configuration.'))

    bootstrap_version = bootstrap_version.split('.')[0]
    if bootstrap_version == '3':
        required_libraries = ('bootstrap-datetimepicker', 'bootstrap-touchspin', 'font-awesome', 'handlebars', 'moment')
    elif bootstrap_version == '4':
        required_libraries = ('bootstrap-touchspin', 'font-awesome', 'handlebars', 'moment', 'tempusdominus')
    else:
        raise HTTPBadRequest(detail=translate('Unsupported Bootstrap version: {0}'.format(bootstrap_version)))

    # Get libraries
    options = dict(
        (option[18:], settings[option])
        for option in settings
        if option.startswith('api_doc.libraries.')
    )

    libraries = []
    for library in options:
        for value in options[library].split():
            if value.startswith(('http://', 'https://', '//')):
                libraries.append(value)
            elif value != 'site':
                libraries.append(request.static_path(value))

    missing = set(required_libraries).difference(options)

    # Embbed CSS and JS files
    view = static_view('pyramid_helpers:static')
    with open(view.find_resource_path('static/css/api-doc-bs{0}.css'.format(bootstrap_version)), 'r') as fp:
        renderer_values['css'] = fp.read()

    with open(view.find_resource_path('static/js/api-doc-bs{0}.js'.format(bootstrap_version)), 'r') as fp:
        renderer_values['js'] = fp.read()

    # Set template according to Bootstrap version
    request.override_renderer = '/api-doc-bs{0}.mako'.format(bootstrap_version)

    # Update renderer values
    renderer_values.update(dict(
        libraries=libraries,
        missing=missing,
        modules=modules,
        orphans=orphans,
    ))

    return renderer_values



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=unused-import,redefined-builtin

""" Form tags for Pyramid """


from pyramid.settings import asbool

#
# Import standard input types
#
from webhelpers2.html.tags import checkbox      # Noqa: F401
from webhelpers2.html.tags import end_form      # Noqa: F401
from webhelpers2.html.tags import file          # Noqa: F401
from webhelpers2.html.tags import form          # Noqa: F401
from webhelpers2.html.tags import hidden        # Noqa: F401
from webhelpers2.html.tags import password      # Noqa: F401
from webhelpers2.html.tags import radio         # Noqa: F401
from webhelpers2.html.tags import select        # Noqa: F401
from webhelpers2.html.tags import submit        # Noqa: F401
from webhelpers2.html.tags import text          # Noqa: F401
from webhelpers2.html.tags import textarea      # Noqa: F401

from webhelpers2.html.tags import OptGroup      # Noqa: F401
from webhelpers2.html.tags import Option        # Noqa: F401
from webhelpers2.html.tags import Options       # Noqa: F401


def coerce_bool(value):
    """ Coerce pseudo boolean to boolean """

    if isinstance(value, str):
        return asbool(value)

    if isinstance(value, bool):
        return value

    raise TypeError('{0!r} could not be coerced to boolean'.format(value))


def html_attrs(**attrs):
    """ Format and clean HTML attributes """

    for key, value in list(attrs.items()):
        if key in ('disabled', 'readonly'):
            if value.lower() not in (key, 'yes', 'true'):
                del attrs[key]
        elif key.startswith('data_'):
            attrs['data-{0}'.format(key[5:])] = attrs.pop(key)

    return attrs


#
# HTML5 new input types
#
def color(name, **attrs):
    """ Create a color field """

    # Override type attribute
    attrs['type'] = 'color'
    return text(name, **attrs)


def date(name, **attrs):
    """ Create a date field """

    # Override type attribute
    attrs['type'] = 'date'
    return text(name, **attrs)


def datetime_local(name, **attrs):
    """ Create a datetime-local field """

    # Override type attribute
    attrs['type'] = 'datetime-local'
    return text(name, **attrs)


def email(name, **attrs):
    """ Create an email field """

    # Override type attribute
    attrs['type'] = 'email'
    return text(name, **attrs)


def month(name, **attrs):
    """ Create a month field """

    # Override type attribute
    attrs['type'] = 'month'
    return text(name, **attrs)


def number(name, **attrs):
    """ Create a number field """

    # Override type attribute
    attrs['type'] = 'number'
    return text(name, **attrs)


def range(name, **attrs):
    """ Create a range field """

    # Override type attribute
    attrs['type'] = 'range'
    return text(name, **attrs)


def search(name, **attrs):
    """ Create a search field """

    # Override type attribute
    attrs['type'] = 'search'
    return text(name, **attrs)


def tel(name, **attrs):
    """ Create a tel field """

    # Override type attribute
    attrs['type'] = 'tel'
    return text(name, **attrs)


def time(name, **attrs):
    """ Create a time field """

    # Override type attribute
    attrs['type'] = 'time'
    return text(name, **attrs)


def url(name, **attrs):
    """ Create an url field """

    # Override type attribute
    attrs['type'] = 'url'
    return text(name, **attrs)


def week(name, **attrs):
    """ Create a week field """

    # Override type attribute
    attrs['type'] = 'week'
    return text(name, **attrs)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP client for Pyramid-Helpers """

import logging
import pprint
import threading
import time

from beaker.cache import Cache
from beaker.util import NoneType
from beaker.util import parse_cache_config_options

import ldap
import ldap.controls
import ldap.modlist

from pyramid_helpers.auth import AuthenticationClient
from pyramid_helpers.utils import ConfigurationError
from pyramid_helpers.utils import get_settings
from pyramid_helpers.utils import parse_settings


SETTINGS_DEFAULTS = {
    'ldap': {
        'tls': False,
        'ca_certs': None,
        'certfile': None,
        'keyfile': None,
        'auth_base_dn': None,
        'auth_filter': 'uid={0}',
        'base_dn': None,
        'bind_dn': None,
        'bind_credential': None,
        'max_retry': 5,
        'retry_after': 300,
        'timeout': 2.0,
    },
}

SETTINGS_RULES = {
    'ldap': [
        ('uri', (str, NoneType), 'uri must be a string designating valid uri'),
        ('tls', (bool, NoneType), 'tls must be a boolean or an integer'),
        ('ca_certs', (str, NoneType), 'ca_certs must be a string designating a filepath'),
        ('certfile', (str, NoneType), 'certfile must be a string designating a filepath'),
        ('keyfile', (str, NoneType), 'keyfile must be a string designating a filepath'),
        ('auth_base_dn', (str, NoneType), 'auth_base_dn must be a string designating a valid dn'),
        ('auth_filter', (str, NoneType), 'auth_filter must be a string designating a valid filter'),
        ('base_dn', (str, NoneType), 'base_dn must be a string designating a valid dn'),
        ('bind_dn', (str, NoneType), 'bind_dn must be a string designating a valid dn'),
        ('bind_credential', (str, NoneType), 'bind_credential must be a string designating a valid credential'),
        ('max_retry', (int, ), 'max_retry must be a string designating a valid integer'),
        ('retry_after', (int, ), 'retry_after must be a string designating a valid integer'),
        ('timeout', (float, ), 'timeout must be a string designating a valid float'),
    ],
}


log = logging.getLogger(__name__)

client = None


class LDAPClient(AuthenticationClient):
    """ LDAP authentication client """

    # pylint: disable=no-member,too-many-instance-attributes

    __auth_mode__ = 'ldap'

    def __init__(self):
        """ LDAP initialization """

        self.uri = None

        self.tls = False
        self.ca_certs = None
        self.certfile = None
        self.keyfile = None

        self.auth_base_dn = None
        self.auth_filter = None
        self.base_dn = None

        self.__bind_dn = None
        self.__bind_credential = None
        self.__cache = None

        self.__conn = None
        self.__lock = threading.Lock()

        # Error handling
        self.max_retry = 5
        self.retry_after = 300
        self.timeout = 2.0

        self.fail_count = 0
        self.wait_until = None

    def __enter__(self):
        """ Acquire lock """

        # pylint: disable=consider-using-with
        self.__lock.acquire()

        return self

    def __exit__(self, type_, value, traceback):
        """ Release lock and close connection """

        self.unbind()
        self.__lock.release()

    def __on_error(self, message, level=logging.ERROR, **kwargs):

        self.fail_count += 1

        if self.fail_count >= self.max_retry:
            self.fail_count = 0
            self.wait_until = time.time() + self.retry_after

            msg = '[LDAP] %s, max failure count was reached (%s), waiting for %s seconds before attempting new connection'
            args = (self.max_retry, self.retry_after)
        else:
            msg = '[LDAP] %s, retry=%s, remaining=%s'
            args = (self.fail_count, self.max_retry - self.fail_count)

        log.log(level, msg, message, *args, **kwargs)

    def add(self, dn, attrs):
        """ Add a new object to LDAP """

        if not self.bind(self.__bind_dn, self.__bind_credential):
            return False

        # Get ldif object suitable for ldap.add()
        ldif = ldap.modlist.addModlist(attrs)
        try:
            # Create LDAP object
            self.__conn.add_s(dn, ldif)
        except ldap.LDAPError:
            log.exception('[LDAP] Failed to create object dn=%s, ldif was:\n%s', dn, pprint.pformat(ldif))
            return False

        log.info('[LDAP] Created object dn=%s.', dn)
        return True

    def bind(self, dn, credential, force=False):
        """ Initialize LDAP connection """

        if not self.__lock.locked():
            raise ldap.LDAPError('Please use LDAPClient inside a `with` statement')

        if self.__conn is not None and not force:
            return True

        if self.wait_until and self.wait_until > time.time():
            return False

        self.wait_until = None

        try:
            self.__conn = ldap.initialize(self.uri)

            self.__conn.set_option(ldap.OPT_NETWORK_TIMEOUT, self.timeout)
            self.__conn.set_option(ldap.OPT_PROTOCOL_VERSION, ldap.VERSION3)
            self.__conn.set_option(ldap.OPT_TIMEOUT, self.timeout)

            if self.ca_certs:
                self.__conn.set_option(ldap.OPT_X_TLS_CACERTFILE, self.ca_certs)

            if self.certfile and self.keyfile:
                self.__conn.set_option(ldap.OPT_X_TLS_CERTFILE, self.certfile)
                self.__conn.set_option(ldap.OPT_X_TLS_KEYFILE, self.keyfile)

            if self.ca_certs or (self.certfile and self.keyfile):
                # Apply pending TLS settings and create a new internal TLS context
                self.__conn.set_option(ldap.OPT_X_TLS_NEWCTX, 0)

            if self.tls:
                self.__conn.start_tls_s()

            if dn and credential:
                self.__conn.simple_bind_s(dn, credential)

        except ldap.INVALID_CREDENTIALS:
            log.error('[LDAP] Authentication failed (%s)', dn)
            return False

        except ldap.LDAPError:
            self.__on_error('Failed to connect to remote server', exc_info=1)
            return False

        # Success
        self.fail_count = 0

        return True

    def get(self, dn, attrlist=None, force=False):
        """ Get an object from it's DN """

        result = self.search(attrlist=attrlist, dn=dn, force=force)
        if not result:
            return None

        return result[0]

    def delete(self, dn):
        """ Delete an object from LDAP """

        if not self.bind(self.__bind_dn, self.__bind_credential):
            return False

        try:
            self.__conn.delete_s(dn)
        except ldap.LDAPError:
            log.exception('[LDAP] Failed to delete object dn=%s.', dn)
            return False

        log.info('[LDAP] Deleted object dn=%s.', dn)
        return True

    def modify(self, dn, old_attrs, new_attrs):
        """ Modify an existing object in LDAP """

        if not self.bind(self.__bind_dn, self.__bind_credential):
            return False

        # Get ldif object suitable for ldap.modify()
        ldif = ldap.modlist.modifyModlist(old_attrs, new_attrs, ignore_attr_types=['objectClass', ])
        if not ldif:
            # No modification
            log.debug('[LDAP] Object dn=%s is uptodate', dn)
            return None

        try:
            self.__conn.modify_s(dn, ldif)
        except ldap.LDAPError:
            log.exception('[LDAP] Failed to modify object dn=%s, ldif was:\n%s', dn, pprint.pformat(ldif))
            return False

        log.info('[LDAP] Modified object dn=%s.', dn)
        return True

    # pylint: disable=too-many-arguments
    def search(self, attrlist=None, base_dn=None, dn=None, filterstr='(objectClass=*)', force=False):
        """ Search entries in LDAP """

        if dn is None:
            if base_dn is None:
                base_dn = self.base_dn
            cache_key = '{0}::{1}'.format(base_dn or '', filterstr)
            scope = ldap.SCOPE_SUBTREE
        else:
            base_dn = dn
            cache_key = dn
            scope = ldap.SCOPE_BASE

        if not force and self.__cache is not None:
            try:
                return self.__cache[cache_key]
            except KeyError:
                pass

        if not self.bind(self.__bind_dn, self.__bind_credential):
            return None

        try:
            result = self.__conn.search_s(base_dn, scope, filterstr=filterstr, attrlist=attrlist)
        except ldap.NO_SUCH_OBJECT:
            return None
        except ldap.LDAPError:
            self.__on_error('Failed to query remote server', exc_info=1)
            return None

        if self.__cache is not None:
            self.__cache[cache_key] = result

        return result

    # pylint: disable=too-many-arguments
    def search_iter(self, attrlist=None, base_dn=None, filterstr='(objectClass=*)', size=100):
        """ Search entries in LDAP (paged) """

        if base_dn is None:
            base_dn = self.base_dn

        if not self.bind(self.__bind_dn, self.__bind_credential):
            return

        control = ldap.controls.SimplePagedResultsControl(True, size=size, cookie='')

        while True:
            try:
                msgid = self.__conn.search_ext(base_dn, ldap.SCOPE_SUBTREE, filterstr=filterstr, attrlist=attrlist, serverctrls=[control])
            except ldap.NO_SUCH_OBJECT:
                return
            except ldap.LDAPError:
                self.__on_error('Failed to query remote server', exc_info=1)
                return

            try:
                _, rdata, _, rcontrols = self.__conn.result3(msgid)
            except ldap.LDAPError:
                self.__on_error('Could not pull LDAP results', exc_info=1)
                break

            # Yield result
            for dn, attrs in rdata:
                yield (dn, attrs)

            # Get page control fron returned controls
            for rcontrol in rcontrols:
                if rcontrol.controlType == ldap.CONTROL_PAGEDRESULTS:
                    break
            else:
                self.__on_error('Server ignores RFC 2696 control')
                break

            # pylint: disable=undefined-loop-variable
            if not rcontrol.cookie:
                # No more result
                break

            control.cookie = rcontrol.cookie

    def setup(self, *args, **kwargs):
        """
        LDAP client setup

        :param uri: URI to connect to
        :param tls: Boolean, enable TLS connectiion
        :param ca_certs: String, path to file with PEM encoded CA certs
        :param certfile: String, path to file with PEM encoded cert for client cert authentication
        :param keyfile: String, path to file with PEM encoded key for client cert authentication
        :param auth_base_dn: Base DN for authentication
        :param auth_filter: Filter string for authentication
        :param base_dn: Base DN for searches
        :param bind_dn: Bind DN to use when querying the server
        :param bind_credentiak: Bind credential to use when querying the server
        :param cache: Cache store
        :param max_retry: Integer, maximum retries before stopping to query the server
        :param retry_after: Integer, number of seconds to wait before retrying to query the server
        :param timeout: Float, timeout value for connections
        """

        self.uri = kwargs.get('uri')

        self.tls = kwargs.get('tls', False)
        self.ca_certs = kwargs.get('ca_certs')
        self.certfile = kwargs.get('certfile')
        self.keyfile = kwargs.get('keyfile')

        self.auth_base_dn = kwargs.get('auth_base_dn')
        self.auth_filter = kwargs.get('auth_filter')
        self.base_dn = kwargs.get('base_dn')

        self.__bind_dn = kwargs.get('bind_dn')
        self.__bind_credential = kwargs.get('bind_credential')
        self.__cache = kwargs.get('cache')

        if kwargs.get('max_retry') is not None:
            self.max_retry = kwargs['max_retry']

        if kwargs.get('retry_after') is not None:
            self.retry_after = kwargs['retry_after']

        if kwargs.get('timeout') is not None:
            self.timeout = kwargs['timeout']

        # Calling inherited
        return super().setup(*args, **kwargs)

    def unbind(self):
        """ Close the connection """

        if self.__conn is None:
            return True

        try:
            self.__conn.unbind()

        except ldap.LDAPError:
            self.__on_error('Failed to unbind connection', level=logging.WARNING)
            return False

        finally:
            self.__conn = None

        return True

    def validate_password(self, username, password, request):
        """ Attempt to bind with the LDAP server using simple authentication """

        if not ldap.dn.is_dn(username):
            if self.auth_filter is None:
                # Configuration error
                log.error('[LDAP] Username must either be dn or auth_filter must be set when calling `.validate_password()`')
                return False

            result = self.search(filterstr=self.auth_filter.format(username), base_dn=self.auth_base_dn)
            if not result:
                # Invalid user
                return False

            username = result[0][0]

        self.unbind()

        return self.bind(username, password)


# Create client object
client = LDAPClient()


def includeme(config):
    """ LDAP module initialization """

    # Load and parse settings
    settings = get_settings(config, 'ldap')
    if settings is None:
        raise ConfigurationError('[LDAP] Invalid or missing configuration for LDAP, please check ldap.filepath directive')

    settings = parse_settings(settings, SETTINGS_RULES, defaults=SETTINGS_DEFAULTS)

    # Connection
    if settings['ldap']['uri'] is None:
        raise ConfigurationError('[LDAP] Missing uri parameter in configuration')

    # Cache setup
    cache_params = parse_cache_config_options(dict(
        ('cache.{0}'.format(k), v)
        for k, v in settings['cache'].items()
    )) if 'cache' in settings else dict(enabled=False)

    if cache_params['enabled']:
        cache = Cache('ldap', **cache_params)
    else:
        cache = None

    # Prepare arguments
    kwargs = settings['ldap'].copy()

    # Setup the client
    client.setup(cache=cache, **kwargs)

    log.info('[LDAP] Initialization complete: uri=%s, cache=%s', client.uri, 'yes, type={type}, expire={expire}'.format(**cache_params) if cache else 'no')



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers articles models """

import datetime

from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column
from sqlalchemy.schema import ForeignKey
from sqlalchemy.types import DateTime
from sqlalchemy.types import Enum
from sqlalchemy.types import Integer
from sqlalchemy.types import Text
from sqlalchemy.types import Unicode

from pyramid_helpers.models import Base
from pyramid_helpers.models import DBSession


class Article(Base):
    """ ORM class mapped to articles table """

    __tablename__ = 'articles'

    # Primary key
    id = Column(Integer, primary_key=True)

    # Foreign key
    author_id = Column(Integer, ForeignKey('users.id'))

    # Attributes
    creation_date = Column(DateTime, nullable=False)
    modification_date = Column(DateTime, nullable=False)
    title = Column(Unicode(255), unique=True, nullable=False)
    text = Column(Text)
    status = Column(Enum('draft', 'published', 'refused'))

    # Relations
    author = relationship('User', backref='articles')

    def from_dict(self, data):
        """ Load data from dict """

        utcnow = datetime.datetime.utcnow()

        if 'author' in data:
            self.author = data['author']

        if 'status' in data:
            self.status = data['status']

        if 'text' in data:
            self.text = data['text']

        if 'title' in data:
            self.title = data['title']

        if self.creation_date is None:
            self.creation_date = data.get('creation_date') or utcnow

        if data.get('modification_date'):
            self.modification_date = data['modification_date']

        elif DBSession.is_modified(self):
            self.modification_date = utcnow

        return DBSession.is_modified(self)

    def to_dict(self, context=None):
        """ Dump data to dict """

        data = dict(
            id=self.id,
            author=self.author,
            title=self.title,
            status=self.status,
        )

        if context == 'api.search':
            data.update(dict(
                creation_date=self.creation_date,
                modification_date=self.modification_date,
            ))
            return data

        data.update(dict(
            text=self.text,
        ))

        return data

    def to_json(self, context=None):
        """ Dump data to JSON """

        data = self.to_dict(context=context)

        data['author'] = self.author.to_json(context=context)
        data['creation_date'] = self.creation_date.isoformat()

        return data



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Articles API
"""

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.httpexceptions import HTTPNotFound
from pyramid.view import view_config

from pyramid_helpers.forms import validate
from pyramid_helpers.forms.articles import CreateForm
from pyramid_helpers.forms.articles import SearchForm
from pyramid_helpers.forms.articles import StatusForm
from pyramid_helpers.funcs.articles import build_search_query
from pyramid_helpers.funcs.articles import create_or_modify
from pyramid_helpers.funcs.articles import get_article
from pyramid_helpers.models import DBSession
from pyramid_helpers.paginate import paginate


@view_config(route_name='api.articles.create', renderer='json', permission='articles.create')
@validate('article', CreateForm)
def create(request):
    """
    Creates an article
    """
    form = request.forms['article']

    article = create_or_modify(request, form)
    if article is None:
        raise HTTPBadRequest(json=form.errors)

    result = dict(
        url=request.url,
        method=request.method,
        params=form.params,
        apiVersion='1.0',
        result=True,
        article=article.to_json(),
    )

    request.response.status_int = 201
    return result


@view_config(route_name='api.articles.delete', renderer='json', permission='articles.delete')
def delete(request):
    """
    Deletes an article
    """
    translate = request.translate

    article = get_article(request)
    if article is None:
        raise HTTPNotFound(detail=translate('Invalid article'))

    DBSession.delete(article)

    result = dict(
        url=request.url,
        method=request.method,
        apiVersion='1.0',
        result=True,
    )

    return result


@view_config(route_name='api.articles.modify', renderer='json', permission='articles.modify')
@validate('article', CreateForm)
def modify(request):
    """
    Modifies an article
    """
    translate = request.translate
    form = request.forms['article']

    article = get_article(request)
    if article is None:
        raise HTTPNotFound(detail=translate('Invalid article'))

    if not create_or_modify(request, form, article=article):
        raise HTTPBadRequest(json=form.errors)

    result = dict(
        url=request.url,
        method=request.method,
        params=form.params,
        apiVersion='1.0',
        result=True,
        article=article.to_json(),
    )

    return result


@view_config(route_name='api.articles.status', renderer='json', permission='articles.modify')
@validate('status', StatusForm)
def status(request):
    """
    Changes an article status
    """
    translate = request.translate
    form = request.forms['status']

    article = get_article(request)
    if article is None:
        raise HTTPNotFound(detail=translate('Invalid article'))

    if form.errors:
        raise HTTPBadRequest(json=form.errors)

    # Update object
    article.from_dict(form.result)

    result = dict(
        url=request.url,
        method=request.method,
        params=form.params,
        apiVersion='1.0',
        result=True,
        article=article.to_json(),
    )

    return result


@view_config(route_name='api.articles.search', renderer='json', permission='articles.search')
@paginate('articles', limit=10, sort='id', order='desc')
@validate('search', SearchForm, method='get')
def search(request):
    """
    Searches in articles
    """
    form = request.forms['search']
    pager = request.pagers['articles']

    if form.errors:
        raise HTTPBadRequest(json=form.errors)

    query = build_search_query(order=pager.order, sort=pager.sort, **form.result)
    pager.set_collection(query)

    result = dict(
        url=request.url,
        method=request.method,
        params=form.params,
        apiVersion='1.0',
        result=True,
        articles=dict(
            items=[article.to_json(context='api.search') for article in pager],
            pager=pager.to_json(),
        )
    )

    return result


@view_config(route_name='api.articles.visual', renderer='json', permission='articles.visual')
def visual(request):
    """
    Gets an article
    """
    translate = request.translate

    article = get_article(request)
    if article is None:
        raise HTTPNotFound(detail=translate('Invalid article'))

    result = dict(
        url=request.url,
        method=request.method,
        apiVersion='1.0',
        result=True,
        article=article.to_json(),
    )

    return result



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Utils functions for Pyramid """

from configparser import ConfigParser
from configparser import Error
import importlib
import logging
import os
import re
import secrets
import string
import warnings

from beaker.util import verify_rules
from decorator import decorator
import pytz

from pyramid.request import Request
from pyramid.request import RequestLocalCache


RANDOM_STRING = string.ascii_letters + string.digits
TIMEZONE = 'Europe/Paris'


log = logging.getLogger(__name__)


class ConfigurationError(Exception):
    """ Configuration error """


def deprecated(old_name, new_func, new_name=None):
    """ Display a deprecation warning when function is called """

    def wrapper(*args, **kwargs):
        warnings.warn(
            '`{0}()` is deprecated and will be removed soon; use `{1}()'.format(wrapper.__old_name__, wrapper.__new_name__),
            DeprecationWarning,
            stacklevel=2,
        )
        return wrapper.__new_func__(*args, **kwargs)

    wrapper.__old_name__ = old_name
    wrapper.__new_func__ = new_func
    wrapper.__new_name__ = new_name or new_func.__name__

    return wrapper


def request_cache():
    """
    Cache decorator

    Values are cached during the lifecycle of a request with a `RequestLocalCache()` instance.

    Warning: The store key is computed from `Request` instance **only**. Other arguments are transparent for caching operation.
    """

    cache = RequestLocalCache()

    def wrapper(func, *args, **kwargs):
        """ Decorator Wrapper function """

        # Get request object from args
        # First arg is «self» when func is a method
        request = None
        for arg in args:
            if isinstance(arg, Request):
                request = arg
                break

        if request is None:
            raise ValueError('Missing request object in args list when calling @request_cache')

        # pylint: disable=unused-argument
        def creator(request):
            func.cache = cache
            return func(*args, **kwargs)

        return cache.get_or_create(request, creator=creator)

    return decorator(wrapper)


def on_before_renderer(event):
    """ Add utc<->local datetime converters to context """

    request = event['request']

    event['localtoutc'] = request.localtoutc
    event['utctolocal'] = request.utctolocal


def on_new_request(event):
    """ Add utc<->local datetime converters to request """

    request = event.request

    def localtoutc(dt, request=request):
        tzinfo = get_tzinfo(request)
        offset = tzinfo.utcoffset(dt)
        return dt - offset

    request.localtoutc = localtoutc

    def utctolocal(dt, request=request):
        tzinfo = get_tzinfo(request)
        return tzinfo.fromutc(dt)

    request.utctolocal = utctolocal


def get_settings(obj, key, section=None):
    """
    Get settings from key
    If key is missing, try to load settings from `key.filepath` INI file
    """

    registry = obj.registry
    settings = registry.settings

    if key not in settings:
        # Load settings from file
        filepath = settings.get('{0}.filepath'.format(key))
        if filepath is None:
            log.debug('Missing %s.filepath directive, please check configuration file', key)
            return None

        filepath = os.path.abspath(filepath)
        if not os.path.isfile(filepath):
            log.error('Invalid %s.filepath directive, please check configuration file', key)
            return None

        if not os.access(filepath, os.R_OK):
            log.error('Not enough permission to access file %s', filepath)
            return None

        parser = ConfigParser(defaults=dict(here=os.path.dirname(filepath)))
        parser.optionxform = str    # Be case sensitive

        try:
            with open(filepath, 'r') as fp:
                parser.read_file(fp)

            settings[key] = dict(
                (section, dict(parser.items(section)))
                for section in parser.sections()
            )
        except (IOError, Error):
            log.exception('Failed to read file %s', filepath)
            return None

    if section is None:
        return settings[key]

    return settings[key].get(section) or {}


def parse_settings(settings, rules, defaults=None):
    """ Parse and check settings """

    if defaults:
        for regex, options in defaults.items():
            found = False
            for section in settings:
                if not re.match(f'^{regex}$', section):
                    continue

                found = True

                for option, value in options.items():
                    settings[section].setdefault(option, value)

            if found or ':' in regex:
                continue

            settings[regex] = options.copy()

    for section in settings:
        settings[section].pop('here', None)

        for regex, rules_data in rules.items():
            if re.match(f'^{regex}$', section):
                verify_rules(settings[section], rules_data)
                break

    return settings


@request_cache()
def get_tzinfo(request):
    """ Get timezone object """

    # Try to get timezone from authenticated user
    if getattr(request, 'authenticated_user', None) is not None:
        timezone = getattr(request.authenticated_user, 'timezone', None)
    else:
        timezone = None

    if timezone is None or timezone not in pytz.common_timezones:
        # Falling back to timezone from settings
        registry = request.registry
        settings = registry.settings

        timezone = settings.get('timezone')

    if timezone is None or timezone not in pytz.common_timezones:
        # Use default timezone
        timezone = TIMEZONE

    return pytz.timezone(timezone)


def random_string(length=10):
    """ Get a random string """

    return ''.join(
        secrets.choice(RANDOM_STRING)
        for _ in range(length)
    )


def resolve_dotted(dotted):
    """ Resolve dotted string to module attribute """

    if not dotted:
        return None

    module_name, module_attr = dotted.rsplit('.', 1)
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        return None

    return getattr(module, module_attr)


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.utils')
    """

    # Subscribers setup
    config.add_subscriber(on_before_renderer, 'pyramid.events.BeforeRender')
    config.add_subscriber(on_new_request, 'pyramid.events.NewRequest')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyramid_helpers',
 'pyramid_helpers.forms',
 'pyramid_helpers.funcs',
 'pyramid_helpers.models',
 'pyramid_helpers.scripts',
 'pyramid_helpers.views',
 'pyramid_helpers.views.api']

package_data = \
{'': ['*'],
 'pyramid_helpers': ['locale/fr/LC_MESSAGES/pyramid-helpers.mo',
                     'static/css/*',
                     'static/js/*',
                     'static/lib/*',
                     'static/lib/bootstrap-3.3.2-dist/css/*',
                     'static/lib/bootstrap-3.3.2-dist/fonts/*',
                     'static/lib/bootstrap-3.3.2-dist/js/*',
                     'static/translations/fr.js',
                     'templates/*',
                     'templates/articles/*']}

install_requires = \
['babel>=2.9,<3.0',
 'colorlog>=5.0,<6.0',
 'decorator>=5.0,<6.0',
 'formencode>=2.0,<3.0',
 'mako>=1.1,<2.0',
 'passlib>=1.7,<2.0',
 'pyramid-beaker>=0.8,<0.9',
 'pyramid-exclog>=1.0,<2.0',
 'pyramid-mako>=1.1,<2.0',
 'pyramid-tm>=2.4,<3.0',
 'pyramid>=2.0,<3.0',
 'pytz',
 'transaction>=3.0,<4.0',
 'webhelpers2>=2.0,<3.0',
 'webob>=1.8,<2.0']

extras_require = \
{'api-doc': ['markdown>=3.3,<4.0'],
 'auth-ldap': ['python-ldap>=3.3,<4.0'],
 'auth-radius': ['pyrad>=2.3,<3.0']}

entry_points = \
{'console_scripts': ['phelpers-init-db = '
                     'pyramid_helpers.scripts.initializedb:main'],
 'paste.app_factory': ['main = pyramid_helpers:main']}

setup_kwargs = {
    'name': 'pyramid-helpers',
    'version': '1.6.2',
    'description': 'Helpers to develop Pyramid applications',
    'long_description': '# Pyramid-Helpers\n\nPyramid-Helpers is a set of helpers to develop applications using Pyramid framework.\n\nIt includes authentication, forms, i18n and pagination helpers.\n\n\n## Prerequisites\nThe project is managed using [Poetry](https://poetry.eustace.io/docs/#installation)\n\n### PostgreSQL adapter (Optional)\nIn order to use a PostgreSQL database, it is recommended to install the [psycopg](https://www.psycopg.org/) adapter. You should check the [build prerequisites](https://www.psycopg.org/docs/install.html#build-prerequisites) in order to install this package (source only).\n\n### LDAP client (Optional)\nLDAP client relies on the [python-ldap](https://www.python-ldap.org/en/latest/) client. You should check the [build prerequisites](https://www.python-ldap.org/en/latest/installing.html#build-prerequisites) in order to install this package.\n\n\n## Development\n```\n# Create virtualenv\nmkdir .venv\npython3 -m venv .venv\n\n# Activate virtualenv\nsource .venv/bin/activate\n\n# Update virtualenv\npip install -U pip setuptools\n\n# Install Poetry\npip install wheel\npip install poetry\n\n# Install application in development mode\npoetry install --extras "{source|binary} [api-doc] [auth-ldap] [auth-radius]"\npoetry run invoke i18n.generate\n\n# Copy and adapt conf/ directory\ncp -a conf .conf\n\n# Initialize database\nphelpers-init-db .conf/application.ini\n\n# Run web server in development mode\npoetry run invoke service.httpd --config-uri=.conf/application.ini --env=.conf/environment\n\n# Run static and functional tests:\npoetry run invoke test\n```\n\n## Tests\n### Static code validation\n```\n# flake8\npoetry run invoke test.flake8\n\n# pylint\npoetry run invoke test.pylint\n\n# both\npoetry run invoke test.static\n```\n\n### Validators\n```\npoetry run invoke test.functional --test=\'tests/test_validators.py\'\n```\n\n### API\n```\npoetry run invoke test.functional --test=\'tests/api/test_articles.py\'\n```\n\n[API tests](/tests/api/test_articles.py) **must** be executed in the correct order, moreover **most** of them depends on previous one.\n\n\n## I18n\nExtract messages\n```\npoetry run invoke i18n.extract i18n.update\n```\n\nCompile catalogs and update JSON files\n```\npoetry run invoke i18n.generate\n```\n\nCreate new language\n```\npoetry run invoke i18n.init {locale_name}\n```\n\n\n## Installation\n\n```\npip install pyramid-helpers\n\n# And optionally:\nphelpers-init-db conf/application.ini\n```\n\n\n## Files list\n\n```\n.\n├── babel.cfg                           Babel configuration file (i18n)\n├── CHANGES.md\n├── pylintrc                            Pylint configuration file\n├── pyproject.toml                      Poetry configuration file\n├── README.md\n├── setup.cfg\n├── conf\n│\xa0\xa0 ├── application.ini                 Main configuration file\n│\xa0\xa0 ├── auth.ini                        Authentication configuration\n│\xa0\xa0 ├── ldap.ini                        LDAP configuration file (auth)\n│   └── radius.ini                      RADIUS configuration file (auth)\n├── pyramid_helpers\n│\xa0\xa0 ├── __init__.py                     demo app config\n│\xa0\xa0 ├── auth.py                         auth helpers for demo app\n│\xa0\xa0 ├── models.py                       SQLAlchemy model for demo app\n│\xa0\xa0 ├── paginate.py                     pagination class, decorator and setup\n│\xa0\xa0 ├── predicates.py                   custom route predicates (Enum, Numeric)\n│\xa0\xa0 ├── resources.py                    basic resource file for demo app\n│\xa0\xa0 ├── forms\n│\xa0\xa0 │\xa0\xa0 ├── __init__.py                 form class, decorator and setup, largely inspired from formhelpers[1]\n│\xa0\xa0 │\xa0\xa0 ├── auth.py                     formencode schema for authentication for demo app\n│\xa0\xa0 │\xa0\xa0 ├── articles.py                 formencode schemas for articles for demo app\n│\xa0\xa0 │\xa0\xa0 ├── tags.py                     HTML input tags renderers\n│\xa0\xa0 │\xa0\xa0 └── validators.py               various formencode validators\n│\xa0\xa0 ├── funcs\n│\xa0\xa0 │\xa0\xa0 ├── __init__.py\n│\xa0\xa0 │\xa0\xa0 └── articles.py                 functions for articles management\n│\xa0\xa0 ├── i18n.py                         i18n setup and helpers\n│\xa0\xa0 ├── locale\n│\xa0\xa0 │\xa0\xa0 ├── fr\n│\xa0\xa0 │\xa0\xa0 │\xa0\xa0 └── LC_MESSAGES\n│\xa0\xa0 │\xa0\xa0 │\xa0\xa0     └── pyramid-helpers.po\n│\xa0\xa0 │\xa0\xa0 └── pyramid-helpers.pot\n│\xa0\xa0 ├── scripts\n│\xa0\xa0 │\xa0\xa0 ├── __init__.py\n│\xa0\xa0 │\xa0\xa0 └── initializedb.py\n│\xa0\xa0 ├── static\n│\xa0\xa0 │\xa0\xa0 ├── css\n│\xa0\xa0 │\xa0\xa0 │\xa0\xa0 └── pyramid-helpers.css     demo app stylesheet\n│\xa0\xa0 │\xa0\xa0 └── js\n│\xa0\xa0 │\xa0\xa0  \xa0\xa0 └── pyramid-helpers.js      demo app javascript\n│\xa0\xa0 ├── templates                       Mako templates\n│\xa0\xa0 │\xa0\xa0 ├── articles                    Mako templates for demo app\n│\xa0\xa0 │\xa0\xa0 │\xa0\xa0 ├── edit.mako\n│\xa0\xa0 │\xa0\xa0 │\xa0\xa0 ├── index.mako\n│\xa0\xa0 │\xa0\xa0 │\xa0\xa0 └── view.mako\n│\xa0\xa0 │\xa0\xa0 ├── confirm.mako\n│\xa0\xa0 │\xa0\xa0 ├── errors.mako\n│\xa0\xa0 │\xa0\xa0 ├── form-tags.mako              Mako templates for forms rendering - derivates from formhelpers[1]\n│\xa0\xa0 │\xa0\xa0 ├── login.mako\n│\xa0\xa0 │\xa0\xa0 ├── paginate.mako               Mako templates for pagination rendering\n│\xa0\xa0 │\xa0\xa0 ├── site.mako                   Main template\n│\xa0\xa0 │\xa0\xa0 └── validators.mako             Test template for validators\n│\xa0\xa0 └── views\n│\xa0\xa0     ├── __init__.py\n│\xa0\xa0     └── articles.py\n├── tasks                               Invoke tasks\n│ \xa0 ├── __init__.py                     initialization\n│ \xa0 ├── common.py                       common file\n│ \xa0 ├── i18n.py                         i18n tasks\n│ \xa0 ├── service.py                      service tasks\n│ \xa0 └── test.py                         test tasks\n└── tests                               Functional tests (pytest)\n    ├── api\n    │\xa0\xa0 └── test_articles.py            test functions for articles API\n    ├── conftest.py                     configuration file for pytest\n    └── test_validators.py              test functions for formencode validators\n```\n\n\n## Useful documentation\n\n* https://docs.pylonsproject.org/projects/pyramid/en/latest/\n* https://docs.pylonsproject.org/projects/pyramid/en/latest/#api-documentation\n* https://techspot.zzzeek.org/2008/07/01/better-form-generation-with-mako-and-pylons/\n',
    'author': 'Cyril Lacoux',
    'author_email': 'clacoux@easter-eggs.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/yack/pyramid-helpers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
